import torch
import torch.nn as nn
import torch.nn.functional as F

from ..attack import Attack


class Jitter(Attack):
    r"""
    Jitter in the paper 'Exploring Misclassifications of Robust Neural Networks to Enhance Adversarial Attacks'
    [https://arxiv.org/abs/2105.10304]

    Distance Measure : Linf

    Arguments:
        model (nn.Module): model to attack.
        eps (float): maximum perturbation. (Default: 8/255)
        alpha (float): step size. (Default: 2/255)
        steps (int): number of steps. (Default: 10)
        random_start (bool): using random initialization of delta. (Default: True)

    Shape:
        - images: :math:`(N, C, H, W)` where `N = number of batches`, `C = number of channels`,        `H = height` and `W = width`. It must have a range [0, 1].
        - labels: :math:`(N)` where each value :math:`y_i` is :math:`0 \leq y_i \leq` `number of labels`.
        - output: :math:`(N, C, H, W)`.

    Examples::
        >>> attack = torchattacks.Jitter(model, eps=8/255, alpha=2/255, steps=10,
                 scale=10, std=0.1, random_start=True)
        >>> adv_images = attack(images, labels)

    """

    def __init__(
        self,
        model,
        eps=8 / 255,
        alpha=2 / 255,
        steps=10,
        scale=10,
        std=0.1,
        random_start=True,
    ):
        super().__init__("Jitter", model)
        self.eps = eps
        self.alpha = alpha
        self.steps = steps
        self.random_start = random_start
        self.scale = scale
        self.std = std
        self.supported_mode = ["default", "targeted"]

    def forward(self, images, labels):
        r"""
        Overridden.
        """

        images = images.clone().detach().to(self.device)
        labels = labels.clone().detach().to(self.device)

        if self.targeted:
            target_labels = self.get_target_label(images, labels)

        loss = nn.MSELoss(reduction="none")

        adv_images = images.clone().detach()

        if self.random_start:
            # Starting at a uniformly random point
            adv_images = adv_images + torch.empty_like(adv_images).uniform_(
                -self.eps, self.eps
            )
            adv_images = torch.clamp(adv_images, min=0, max=1).detach()

        for _ in range(self.steps):
            adv_images.requires_grad = True
            logits = self.get_logits(adv_images)

            _, pre = torch.max(logits, dim=1)
            wrong = pre != labels

            norm_z = torch.norm(logits, p=float("inf"), dim=1, keepdim=True)
            hat_z = nn.Softmax(dim=1)(self.scale * logits / norm_z)

            if self.std != 0:
                hat_z = hat_z + self.std * torch.randn_like(hat_z)

            # Calculate loss
            if self.targeted:
                target_Y = F.one_hot(
                    target_labels, num_classes=logits.shape[-1]
                ).float()
                cost = -loss(hat_z, target_Y).mean(dim=1)
            else:
                Y = F.one_hot(labels, num_classes=logits.shape[-1]).float()
                cost = loss(hat_z, Y).mean(dim=1)

            norm_r = torch.norm(
                (adv_images - images), p=float("inf"), dim=[1, 2, 3]
            )  # nopep8
            nonzero_r = norm_r != 0
            cost[wrong * nonzero_r] /= norm_r[wrong * nonzero_r]

            cost = cost.mean()

            # Update adversarial images
            grad = torch.autograd.grad(
                cost, adv_images, retain_graph=False, create_graph=False
            )[0]

            adv_images = adv_images.detach() + self.alpha * grad.sign()
            delta = torch.clamp(
                adv_images - images, min=-self.eps, max=self.eps
            )  # nopep8
            adv_images = torch.clamp(images + delta, min=0, max=1).detach()

        return adv_images



import torch
import torch.nn as nn

from ..attack import Attack


class PGD(Attack):
    r"""
    PGD in the paper 'Towards Deep Learning Models Resistant to Adversarial Attacks'
    [https://arxiv.org/abs/1706.06083]

    Distance Measure : Linf

    Arguments:
        model (nn.Module): model to attack.
        eps (float): maximum perturbation. (Default: 8/255)
        alpha (float): step size. (Default: 2/255)
        steps (int): number of steps. (Default: 10)
        random_start (bool): using random initialization of delta. (Default: True)

    Shape:
        - images: :math:`(N, C, H, W)` where `N = number of batches`, `C = number of channels`,        `H = height` and `W = width`. It must have a range [0, 1].
        - labels: :math:`(N)` where each value :math:`y_i` is :math:`0 \leq y_i \leq` `number of labels`.
        - output: :math:`(N, C, H, W)`.

    Examples::
        >>> attack = torchattacks.PGD(model, eps=8/255, alpha=1/255, steps=10, random_start=True)
        >>> adv_images = attack(images, labels)

    """

    def __init__(self, model, eps=8 / 255, alpha=2 / 255, steps=10, random_start=True):
        super().__init__("PGD", model)
        self.eps = eps
        self.alpha = alpha
        self.steps = steps
        self.random_start = random_start
        self.supported_mode = ["default", "targeted"]

    def forward(self, images, labels):
        r"""
        Overridden.
        """

        images = images.clone().detach().to(self.device)
        labels = labels.clone().detach().to(self.device)

        if self.targeted:
            target_labels = self.get_target_label(images, labels)

        loss = nn.CrossEntropyLoss()
        adv_images = images.clone().detach()

        if self.random_start:
            # Starting at a uniformly random point
            adv_images = adv_images + torch.empty_like(adv_images).uniform_(
                -self.eps, self.eps
            )
            adv_images = torch.clamp(adv_images, min=0, max=1).detach()

        for _ in range(self.steps):
            adv_images.requires_grad = True
            outputs = self.get_logits(adv_images)

            # Calculate loss
            if self.targeted:
                cost = -loss(outputs, target_labels)
            else:
                cost = loss(outputs, labels)

            # Update adversarial images
            grad = torch.autograd.grad(
                cost, adv_images, retain_graph=False, create_graph=False
            )[0]

            adv_images = adv_images.detach() + self.alpha * grad.sign()
            delta = torch.clamp(adv_images - images, min=-self.eps, max=self.eps)
            adv_images = torch.clamp(images + delta, min=0, max=1).detach()

        return adv_images



# (c) 2024 LAiSR-SK
# This code is licensed under the MIT license (see LICENSE.md).
import numpy as np
import torch
from autoattack.autopgd_base import APGDAttack

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")


class APGD:
    """
    APGD attack (from AutoAttack) (Croce et al, 2020).
    The attack performs nb_iter steps of adaptive size, while always staying within eps from the initial point.
    Arguments:
        predict (nn.Module): forward pass function.
        loss_fn (str): loss function - ce or dlr.
        n_restarts (int): number of random restarts.
        eps (float): maximum distortion.
        nb_iter (int): number of iterations.
        ord (int): (optional) the order of maximum distortion (inf or 2).
    """

    def __init__(
        self,
        predict,
        loss_fn="ce",
        n_restarts=2,
        eps=0.3,
        nb_iter=40,
        ord=np.inf,
        seed=1,
    ):
        assert loss_fn in [
            "ce",
            "dlr",
        ], "Only loss_fn=ce or loss_fn=dlr are supported!"
        assert ord in [2, np.inf], "Only ord=inf or ord=2 are supported!"

        norm = "Linf" if ord == np.inf else "L2"
        self.apgd = APGDAttack(
            predict,
            n_restarts=n_restarts,
            n_iter=nb_iter,
            verbose=False,
            eps=eps,
            norm=norm,
            eot_iter=1,
            rho=0.75,
            seed=seed,
            device=device,
        )
        self.apgd.loss = loss_fn

    def perturb(self, x, y):
        x_adv = self.apgd.perturb(x, y)[1]
        r_adv = x_adv - x
        return x_adv, r_adv


class LinfAPGDAttack(APGD):
    """
    APGD attack (from AutoAttack) with order=Linf.
    The attack performs nb_iter steps of adaptive size, while always staying within eps from the initial point.
    Arguments:
        predict (nn.Module): forward pass function.
        loss_fn (str): loss function - ce or dlr.
        n_restarts (int): number of random restarts.
        eps (float): maximum distortion.
        nb_iter (int): number of iterations.
    """

    def __init__(
        self, predict, loss_fn="ce", n_restarts=2, eps=0.3, nb_iter=40, seed=1
    ):
        ord = np.inf
        super(L2APGDAttack, self).__init__(
            predict=predict,
            loss_fn=loss_fn,
            n_restarts=n_restarts,
            eps=eps,
            nb_iter=nb_iter,
            ord=ord,
            seed=seed,
        )


class L2APGDAttack(APGD):
    """
    APGD attack (from AutoAttack) with order=L2.
    The attack performs nb_iter steps of adaptive size, while always staying within eps from the initial point.
    Arguments:
        predict (nn.Module): forward pass function.
        loss_fn (str): loss function - ce or dlr.
        n_restarts (int): number of random restarts.
        eps (float): maximum distortion.
        nb_iter (int): number of iterations.
    """

    def __init__(
        self, predict, loss_fn="ce", n_restarts=2, eps=0.3, nb_iter=40, seed=1
    ):
        ord = 2
        super(L2APGDAttack, self).__init__(
            predict=predict,
            loss_fn=loss_fn,
            n_restarts=n_restarts,
            eps=eps,
            nb_iter=nb_iter,
            ord=ord,
            seed=seed,
        )



# Import necessary libraries
import matplotlib.pyplot as plt
import torch.nn as nn
import torch.nn.functional as F
from attacks.attacks.apgd import APGD
from attacks.attacks.apgdt import APGDT
from attacks.attacks.autoattack import AutoAttack
from attacks.attacks.bim import BIM
from attacks.attacks.cw import CW
from attacks.attacks.deepfool import DeepFool
from attacks.attacks.difgsm import DIFGSM
from attacks.attacks.eaden import EADEN
from attacks.attacks.eadl1 import EADL1
from attacks.attacks.eotpgd import EOTPGD
from attacks.attacks.fab import FAB
from attacks.attacks.ffgsm import FFGSM
from attacks.attacks.fgsm import FGSM
from attacks.attacks.gn import GN
from attacks.attacks.jitter import Jitter
from attacks.attacks.jsma import JSMA
from attacks.attacks.mifgsm import MIFGSM
from attacks.attacks.nifgsm import NIFGSM
from attacks.attacks.onepixel import OnePixel
from attacks.attacks.pgd import PGD
from attacks.attacks.pgdl2 import PGDL2
from attacks.attacks.pgdrs import PGDRS
from attacks.attacks.pgdrsl2 import PGDRSL2
from attacks.attacks.pifgsm import PIFGSM
from attacks.attacks.pifgsmpp import PIFGSMPP
from attacks.attacks.pixle import Pixle
from attacks.attacks.rfgsm import RFGSM
from attacks.attacks.sinifgsm import SINIFGSM
from attacks.attacks.sparsefool import SparseFool
from attacks.attacks.spsa import SPSA
from attacks.attacks.square import Square
from attacks.attacks.tifgsm import TIFGSM
from attacks.attacks.tpgd import TPGD
from attacks.attacks.upgd import UPGD
from attacks.attacks.vanila import VANILA
from attacks.attacks.vmifgsm import VMIFGSM
from attacks.attacks.vnifgsm import VNIFGSM
from torch.utils.data import DataLoader
from torchvision import datasets, transforms


# Define a simple CNN model
class Net(nn.Module):
    def __init__(self):
        super(Net, self).__init__()
        self.conv1 = nn.Conv2d(1, 10, kernel_size=5)
        self.conv2 = nn.Conv2d(10, 20, kernel_size=5)
        self.conv2_drop = nn.Dropout2d()
        self.fc1 = nn.Linear(320, 50)
        self.fc2 = nn.Linear(50, 10)

    def forward(self, x):
        x = F.relu(F.max_pool2d(self.conv1(x), 2))
        x = F.relu(F.max_pool2d(self.conv2_drop(self.conv2(x)), 2))
        x = x.view(-1, 320)
        x = F.relu(self.fc1(x))
        x = F.dropout(x, training=self.training)
        x = self.fc2(x)
        return F.log_softmax(x, dim=1)


model = Net()

# Define the attacks
attacks = [
    VNIFGSM,
    VMIFGSM,
    VANILA,
    UPGD,
    TPGD,
    Square,
    SPSA,
    SparseFool,
    SINIFGSM,
    RFGSM,
    Pixle,
    PGDRSL2,
    PGDRS,
    PGDL2,
    NIFGSM,
    MIFGSM,
    JSMA,
    Jitter,
    GN,
    FFGSM,
    FAB,
    EOTPGD,
    EADL1,
    EADEN,
    DIFGSM,
    BIM,
    AutoAttack,
    APGDT,
    APGD,
    FGSM,
    PGD,
    CW,
    DeepFool,
    OnePixel,
]
attacks2 = [
    TIFGSM,
    PIFGSMPP,
    PIFGSM,
]
# Load the MNIST dataset
mnist = datasets.MNIST(
    root="./data", train=True, download=True, transform=transforms.ToTensor()
)
mnist_loader = DataLoader(mnist, batch_size=1, shuffle=True)

# Get a single image from the MNIST dataset
dataiter = iter(mnist_loader)
images, labels = next(dataiter)

adversarial_examples = {}

for attack in attacks:
    attacker = attack(model)
    adversarial_example = attacker(images, labels)
    adversarial_examples[attack.__name__] = adversarial_example


fig, axes = plt.subplots(1, len(adversarial_examples))

for i, (attack_name, adversarial_example) in enumerate(
    adversarial_examples.items()
):
    adversarial_example_np = adversarial_example.detach().cpu().numpy()

    # If the image has 4 dimensions, take the first dimension (assuming it's a single image)
    if len(adversarial_example_np.shape) == 4:
        adversarial_example_np = adversarial_example_np[0, 0]

    # Plot the adversarial example in the current subplot
    axes[i].imshow(adversarial_example_np, cmap="gray")
    axes[i].axis("off")

# Show the plot
plt.show()



import numpy as np

import torch
import torch.nn.functional as F

from ..attack import Attack
from ._differential_evolution import differential_evolution


class OnePixel(Attack):
    r"""
    Attack in the paper 'One pixel attack for fooling deep neural networks'
    [https://arxiv.org/abs/1710.08864]

    Modified from "https://github.com/DebangLi/one-pixel-attack-pytorch/" and 
    "https://github.com/sarathknv/adversarial-examples-pytorch/blob/master/one_pixel_attack/"

    Distance Measure : L0

    Arguments:
        model (nn.Module): model to attack.
        pixels (int): number of pixels to change (Default: 1)
        steps (int): number of steps. (Default: 10)
        popsize (int): population size, i.e. the number of candidate agents or "parents" in differential evolution (Default: 10)
        inf_batch (int): maximum batch size during inference (Default: 128)

    Shape:
        - images: :math:`(N, C, H, W)` where `N = number of batches`, `C = number of channels`,        `H = height` and `W = width`. It must have a range [0, 1].
        - labels: :math:`(N)` where each value :math:`y_i` is :math:`0 \leq y_i \leq` `number of labels`.
        - output: :math:`(N, C, H, W)`.

    Examples::
        >>> attack = torchattacks.OnePixel(model, pixels=1, steps=10, popsize=10, inf_batch=128)
        >>> adv_images = attack(images, labels)

    """

    def __init__(self, model, pixels=1, steps=10, popsize=10, inf_batch=128):
        super().__init__("OnePixel", model)
        self.pixels = pixels
        self.steps = steps
        self.popsize = popsize
        self.inf_batch = inf_batch
        self.supported_mode = ["default", "targeted"]

    def forward(self, images, labels):
        r"""
        Overridden.
        """

        images = images.clone().detach().to(self.device)
        labels = labels.clone().detach().to(self.device)

        if self.targeted:
            target_labels = self.get_target_label(images, labels)

        batch_size, channel, height, width = images.shape

        bounds = [(0, height), (0, width)] + [(0, 1)] * channel
        bounds = bounds * self.pixels

        popmul = max(1, int(self.popsize / len(bounds)))

        adv_images = []
        for idx in range(batch_size):
            image, label = images[idx : idx + 1], labels[idx : idx + 1]

            if self.targeted:
                target_label = target_labels[idx : idx + 1]

                def func(delta):
                    return self._loss(image, target_label, delta)

                def callback(delta, convergence):
                    return self._attack_success(image, target_label, delta)

            else:

                def func(delta):
                    return self._loss(image, label, delta)

                def callback(delta, convergence):
                    return self._attack_success(image, label, delta)

            delta = differential_evolution(
                func=func,
                bounds=bounds,
                callback=callback,
                maxiter=self.steps,
                popsize=popmul,
                init="random",
                recombination=1,
                atol=-1,
                polish=False,
            ).x
            delta = np.split(delta, len(delta) / len(bounds))
            adv_image = self._perturb(image, delta)
            adv_images.append(adv_image)

        adv_images = torch.cat(adv_images)
        return adv_images

    def _loss(self, image, label, delta):
        adv_images = self._perturb(image, delta)  # Mutiple delta
        prob = self._get_prob(adv_images)[:, label]
        if self.targeted:
            return 1 - prob  # If targeted, increase prob
        else:
            return prob  # If non-targeted, decrease prob

    def _attack_success(self, image, label, delta):
        adv_image = self._perturb(image, delta)  # Single delta
        prob = self._get_prob(adv_image)
        pre = np.argmax(prob)
        if self.targeted and (pre == label):
            return True
        elif (not self.targeted) and (pre != label):
            return True
        return False

    def _get_prob(self, images):
        with torch.no_grad():
            batches = torch.split(images, self.inf_batch)
            outs = []
            for batch in batches:
                out = self.get_logits(batch)
                outs.append(out)
        outs = torch.cat(outs)
        prob = F.softmax(outs, dim=1)
        return prob.detach().cpu().numpy()

    def _perturb(self, image, delta):
        delta = np.array(delta)
        if len(delta.shape) < 2:
            delta = np.array([delta])
        num_delta = len(delta)
        adv_image = image.clone().detach().to(self.device)
        adv_images = torch.cat([adv_image] * num_delta, dim=0)
        for idx in range(num_delta):
            pixel_info = delta[idx].reshape(self.pixels, -1)
            for pixel in pixel_info:
                pos_x, pos_y = pixel[:2]
                channel_v = pixel[2:]
                for channel, v in enumerate(channel_v):
                    adv_images[idx, channel, int(pos_x), int(pos_y)] = v
        return adv_images



# Modified from https://github.com/bearpaw/pytorch-classification/blob/master/models/cifar/wrn.py
import torch
import torch.nn as nn


class Identity(nn.Module):
    def forward(self, inputs):
        return inputs



# (c) 2024 LAiSR-SK

# This code is licensed under the MIT license (see LICENSE.md).

"""params must provide: --data, --net, --method, --expect_acc/--validk, --maxk(for cat)"""


import os
from collections import namedtuple
from os import PathLike
from typing import Optional, OrderedDict

import numpy as np
import torch
import torch.backends.cudnn as cudnn
import torchvision
import torchvision.transforms as transforms
from torch import nn, optim
from torch.autograd import Variable
from torch.utils.data.sampler import SubsetRandomSampler

from airtk.defense.currat.currat_pgd import (
    LinfPGDAttack,
)
from airtk.model import ResNet50, WideResNet

# with open("training/currat/currat_config.json") as config_file:

# config = json.load(config_file)


# seeding randomness

# np.random.seed(config["np_random_seed"])


# Setting up training parameters

# train_batch_size = config["training_batch_size"]

# eval_batch_size = max(64, config["eval_batch_size"])

# criterion = nn.CrossEntropyLoss()


# parser = argparse.ArgumentParser(description="CAT Training")

# args = parser.parse_args()


class CurratTraining:
    def __init__(
        self,
        dataset_name: str,
        model_name: str,
        batch_size: int = 200,
        epochs: int = 300,
        lr: float = 0.1,
        method: str = "cat",
        max_k: int = 0,
        k: int = 1,
        valid_k: Optional[int] = None,
        use_sgd: bool = False,
        model_dir: PathLike = "./data/model",
        expect_acc: float = 0.95,
    ) -> None:
        ArgPrototype = namedtuple(
            "ArgPrototype",
            [
                "lr",
                "resume",
                "resume_from",
                "save_name",
                "batch_size",
                "expect_acc",
                "net",
                "data",
                "method",
                "maxk",
                "k",
                "validk",
                "sgd",
                "epochs",
            ],
        )

        self._args = ArgPrototype(
            lr,
            False,
            -1,
            model_dir,
            batch_size,
            expect_acc,
            model_name,
            dataset_name,
            method,
            max_k,
            k,
            valid_k,
            use_sgd,
            epochs,
        )

        if dataset_name == "mnist":
            self._args = self._args._replace(net="cnn")

    def __call__(self) -> None:
        main_currat(self._args.data, self._args)


# adjust

# if args.data == "mnist":

# args.net == "cnn"

#     args.k=40

# else:

#     args.k=7


# max_k = 40 if args.data == "mnist" else 7

# if args.maxk > 0:

# max_k = args.maxk

# if args.savename is None:

# args.savename = args.method + "." + args.data + "." + args.net

# savefolder = args.savename

# if args.method == "cat":

# savefolder += ".maxk" + str(max_k)

# if args.validk is not None:

# savefolder += ".valid" + args.validk

# elif args.method == "cat" or args.method == "basic":

# savefolder += ".k+%.2f" % args.expect_acc

# if args.sgd:

# savefolder += ".sgd"

# expect_acc = float(args.expect_acc)

#

# use_cuda = torch.cuda.is_available()

# start_epoch = 0  # start from epoch 0 or last checkpoint epoch


class Flatten(nn.Module):
    def forward(self, x):
        return x.view(x.size()[0], -1)


# Model


def resumeFrom(modelFN, begin=True):
    global k, prevTime, optimizer, net

    # Load checkpoint.

    print("==> Resuming from checkpoint..", modelFN)

    assert os.path.exists(modelFN), "Error: no checkpoint file found!"

    if begin:
        prevTime = int(os.path.basename(modelFN).split(".")[5][:-3])

    print("prevTime", prevTime)

    checkpoint = torch.load(modelFN)

    try:
        net.load_state_dict(checkpoint["net"])

    except Exception:
        new_state_dict = OrderedDict()

        for k, v in checkpoint["net"].items():
            name = k[7:]  # remove `module.`

            new_state_dict[name] = v

        net.load_state_dict(new_state_dict)

    this_epoch = checkpoint["epoch"]

    k = checkpoint["k"]

    if os.path.exists(modelFN + ".optim"):
        optimizer = optim.Adam(net.parameters(), lr=1e-4)

        optimizer.load_state_dict(torch.load(modelFN + ".optim")["optim"])

    else:
        pass

    return this_epoch


# k, prevTime = 0, 0

# net = None

# print("==> Building model..")

# # if args.net == "cnn":

# net = nn.Sequential(

# nn.Conv2d(1, 32, 5, stride=1, padding=2),

# nn.ReLU(),

# nn.MaxPool2d(2),

# nn.Conv2d(32, 64, 5, stride=1, padding=2),

# nn.ReLU(),

# nn.MaxPool2d(2),

# Flatten(),

# nn.Linear(64 * 7 * 7, 1024),

# nn.ReLU(),

# nn.Linear(1024, 10),

# )

# elif args.net=='resnet50': net = ResNet50()

# elif args.net=='densenet161': net = DenseNet161()

# elif args.net == "wideresnet":

# net = WideResNet(

# depth=34, num_classes=100 if args.data == "cifar100" else 10

# )

# optimizer = None  # optim.Adam(net.parameters(), lr=1e-4)

# if args.resume:

# if args.resumefrom is None:

# model_dir = args.save_name

# all_models = [

# name

# for name in os.listdir(model_dir)

# if name.split(".")[-1] != "optim"

# ]

# assert len(all_models) >= 1, (

# "Error: no checkpoint file found in " + model_dir

# )

# args.resumefrom = os.path.join(

# model_dir,

# sorted(all_models, key=lambda x: int(x.split(".")[3]))[-1],

# )  # the biggest model

# print("args.resumefrom", args.resumefrom)

# start_epoch = resumeFrom(args.resumefrom)


# if use_cuda:

# net.cuda()

# net = torch.nn.DataParallel(

# net, device_ids=range(torch.cuda.device_count())

# )

# cudnn.benchmark = True

# if optimizer is None:

# optimizer = optim.Adam(net.parameters(), lr=1e-4)

# if args.sgd:

# optimizer = optim.SGD(

# net.parameters(), lr=args.lr, momentum=0.9, weight_decay=5e-4

# )

# epsilon, step_size = 0, 0

# if args.data == "mnist":

# epsilon, step_size = 0.3, 0.6

# elif args.data == "cifar10" or args.data == "cifar100":

# epsilon, step_size = 0.032, 0.064

# elif args.data == "svhn":

# epsilon, step_size = 0.047, 0.094

# attack = LinfPGDAttack(net, epsilon, 1, step_size, True, "xent")


def getKset(curk, args):
    if args.method == "at" or args.method == "maxbm":
        return [1, 20, 100]

    if args.method == "nat":
        return [0]

    if args.validk == "byk":
        if curk <= 10:
            return [i for i in range(curk + 1)]

        elif curk > 10 and curk < 20:
            return [i for i in range(11)]

        elif curk >= 20:
            return [i for i in range(11)] + [20]

    elif args.validk == "k":
        return [curk]


def regularLoss(
    inputs, targets
):  # in: inputs, targets; out: # correct prediction, loss value
    global k, net

    criterion = nn.CrossEntropyLoss()

    inputs_var, targets_var = Variable(inputs), Variable(targets)

    each_batch = int(inputs.size(0) / (k + 1))

    if each_batch == 0:
        return -1, -1

    for cur_k in range(k + 1):
        if cur_k != k:
            inputs_now, targets_now = (
                Variable(
                    inputs[cur_k * each_batch : (cur_k + 1) * each_batch]
                ),
                Variable(
                    targets[cur_k * each_batch : (cur_k + 1) * each_batch]
                ),
            )

        else:
            inputs_now, targets_now = (
                Variable(inputs[k * each_batch :]),
                Variable(targets[k * each_batch :]),
            )  # this gets max share

        inputs_adv = (
            attack.perturb_true(inputs_now, targets_now, cur_k)
            if cur_k != 0
            else inputs_now
        )

        if cur_k != k:
            (
                inputs_var[cur_k * each_batch : (cur_k + 1) * each_batch],
                targets_var[cur_k * each_batch : (cur_k + 1) * each_batch],
            ) = inputs_adv, targets_now

        else:
            inputs_var[k * each_batch :], targets_var[k * each_batch :] = (
                inputs_adv,
                targets_now,
            )

    # img = torchvision.utils.make_grid(inputs_adv.data[-3:,:,:,:])

    outputs = net(inputs_var)

    loss = criterion(outputs, targets_var)

    # print('reg loss', loss.data[0])

    optimizer.zero_grad()

    loss.backward()

    optimizer.step()

    _, predicted = torch.max(outputs.data, 1)

    current_correct = predicted.eq(targets_var.data).cpu().sum()

    return current_correct, loss.item()


prevBestEpoch = {"fn": None, "eid": 0, "acc": 0, "loss": 0}


def advtrain_CAT(
    epoch, trainloader, validloader, args
):  # basic + batch mixing
    global k, max_k, prevBestEpoch, net

    use_cuda = torch.cuda.is_available()

    print("\nEpoch: %d" % epoch, "k: %d" % k)
    net.train()

    train_loss = 0

    correct = 0

    total = 0

    for _, (inputs, targets) in enumerate(trainloader):
        if use_cuda:
            inputs, targets = inputs.cuda(), targets.cuda()

        #########

        current_correct, current_loss = regularLoss(inputs, targets)

        if current_correct == -1 and current_loss == -1:
            continue

        ###############

        correct += current_correct

        train_loss += current_loss

        current_num = Variable(targets).size(0)

        total += current_num

        """progress_bar(batch_idx, len(trainloader), 'Loss: %.3f | Acc: %.3f%% (%d/%d)'

            % (train_loss/(batch_idx+1), 100.*correct/total, correct, total))"""

    acc = 1.0 * correct / total

    return validnSaveModel(acc, k, validloader, args)


def validnSaveModel(
    trainAcc, inK, validloader, args, new=True
):  # , k, acc tbd: saveModel contains: state->time->filename(including acc&k)->...
    global starttime, net, epoch, optimizer, k

    savefolder = args.save_name

    # if not os.path.isdir("checkpoint"):

    # os.mkdir("checkpoint")

    # if not os.path.isdir("checkpoint/" + savefolder):

    # os.mkdir("checkpoint/" + savefolder)

    state = {
        "net": net.state_dict(),  # net.module if use_cuda else net,
        "k": inK,  # could be k or static_k
        "epoch": epoch,
    }

    state_optim = {"optim": optimizer.state_dict()}

    # sofartime = timeit.default_timer()

    savename = (
        # "./checkpoint/"
        savefolder
        + "/"
        # + args.savename
        + "."
        + str(epoch)
        + ".step%d" % (inK)
        + "."
        # + "%.f" % (sofartime - starttime)
    )

    if args.validk is not None:  # use validation
        validAccNQ, validLossNQ, validAccQ, validLossQ = valid_wca(
            getKset(inK, args), validloader
        )

        savename += (
            ".validNQ"
            + "%.1f" % validAccNQ
            + ".lossNQ"
            + "%.1f" % validLossNQ
            + ".validQ"
            + "%.1f" % validAccQ
            + ".lossQ"
            + "%.1f" % validLossQ
            + ".train"
            + "%.1f" % (trainAcc)
        )

        torch.save(state, savename)

        saveOptim(
            state_optim, savename + ".optim", epoch
        )  # torch.save(state_optim, savename_optim)

        # if (validAcc>prevBestEpoch['acc'] or validLoss<prevBestEpoch['loss']) and trainAcc>0.9:
        print(
            "new model:",
            savename,
            "validAccNQ:",
            validAccNQ,
            "validAccQ:",
            validAccQ,
            "trainAcc:",
            trainAcc,
        )

        validAcc = validAccQ if args.method == "cat" else validAccNQ

        validLoss = validLossQ if args.method == "cat" else validLossNQ

        if (validAcc > prevBestEpoch["acc"]) or (not new):
            prevBestEpoch["fn"] = savename

            prevBestEpoch["eid"] = epoch

            prevBestEpoch["acc"] = validAcc

            prevBestEpoch["loss"] = validLoss

            print("new prev best", savename, epoch, validAcc)
        elif (
            epoch - prevBestEpoch["eid"] >= 10 and trainAcc > 0.9
        ):  # resume from prev Epoch
            prevBestEpoch["acc"] = (
                0  # just make sure to save the first epoch in new k
            )

            if args.method == "cat" or args.method == "basic":
                removeFNs(prevBestEpoch["fn"], prevBestEpoch["eid"] + 1, epoch)

                resumeFrom(prevBestEpoch["fn"], False)

                k += 1

                epoch += 1

                print("use prev best model for new k", k)

                validnSaveModel(
                    0, k, validloader, args, False
                )  # recursive call validnSave to validate new k set with previous best model
                print(
                    "cur epoch:",
                    epoch,
                    "next k",
                    k,
                    "prevBestEpoch['eid']",
                    prevBestEpoch["eid"],
                )

            else:
                print(
                    "cur epoch:",
                    epoch,
                    "should have resumed at",
                    prevBestEpoch["eid"],
                )

    else:
        savename += ".train" + "%.1f" % (trainAcc)

        torch.save(state, savename)

        saveOptim(
            state_optim, savename + ".optim", epoch
        )  # torch.save(state_optim, savename_optim)
        if (
            (args.method == "cat" or args.method == "basic")
            and trainAcc > args.expect_acc
            and k < max_k
        ):  # use max acc
            k += 1

    return epoch


def saveOptim(state_optim, savename_optim, epoch):
    torch.save(state_optim, savename_optim)

    tmp = savename_optim.split("/")
    fn = (
        "/".join(tmp[:-1])
        + "/"
        + ".".join(tmp[-1].split(".")[:3])
        + "."
        + str(epoch - 11)
        + ".*"
        + ".optim"
    )

    # print("rm " + fn)

    # try:

    # os.system("rm " + fn)

    # except Exception:

    # pass


def removeFNs(fileTemp, startE, endE):
    print("startE, endE", startE, endE)

    for ie in range(startE, endE + 1):
        tmp = fileTemp.split("/")
        fn = (
            "/".join(tmp[:-1])
            + "/"
            + ".".join(tmp[-1].split(".")[:3])
            + "."
            + str(ie)
            + ".*"
        )

        # print("to delete", fn + ".optim")

        # try:

        # os.system('rm '+fn)

        # os.system("rm " + fn + ".optim")

        # except Exception:

        # pass


def valid_wca(kset, validloader):
    print("kset", kset)

    criterion = nn.CrossEntropyLoss()

    net.eval()

    test_lossNQ, test_lossQ = 0, 0

    totalcorrectNQ, totalcorrectQ = 0, 0

    total = 0

    for batch_idx, (inputs, targets) in enumerate(validloader):
        if torch.cuda.is_available():
            inputs, targets = inputs.cuda(), targets.cuda()

        inputs, targets = Variable(inputs), Variable(targets)

        correctNQ, correctQ = None, None

        for k in kset:
            inputs_advNQ, inputs_advQ = inputs, inputs

            if k > 0:
                inputs_advNQ, inputs_advQ = attack.perturb_quad(
                    inputs, targets, k, 16
                )

            outputsNQ, outputsQ = net(inputs_advNQ), net(inputs_advQ)

            lossNQ = criterion(outputsNQ, targets)

            lossQ = criterion(outputsQ, targets)

            test_lossNQ += lossNQ.item()

            test_lossQ += (
                lossQ.item()
            )  # this one too, switched item() from data

            _, predictedNQ = torch.max(outputsNQ, 1)

            _, predictedQ = torch.max(outputsQ, 1)

            if correctNQ is None:
                correctNQ = predictedNQ.eq(targets.data).cpu()

            else:
                correctNQ *= predictedNQ.eq(targets.data).cpu()

            if correctQ is None:
                correctQ = predictedQ.eq(targets.data).cpu()

            else:
                correctQ *= predictedQ.eq(targets.data).cpu()

        total += targets.size(0)

        totalcorrectNQ += correctNQ.sum()

        totalcorrectQ += correctQ.sum()

        """progress_bar(batch_idx, len(validloader), 'valid_Loss: %.3f | valid_Acc: %.3f%% (%d/%d)'

                     % (test_lossQ / (batch_idx + 1), 100. * totalcorrectQ / total, totalcorrectQ, total))"""

    accNQ = 100.0 * totalcorrectNQ / total

    print("valid acc NQ", accNQ, "valid loss NQ", test_lossNQ)

    accQ = 100.0 * totalcorrectQ / total

    print("valid acc Q", accQ, "valid loss Q", test_lossQ)

    return accNQ, test_lossNQ, accQ, test_lossQ


# starttime = timeit.default_timer() - prevTime

# for epoch in range(start_epoch, start_epoch+300):

# epoch = start_epoch + 1


def main_currat(ds_name, args):
    global starttime, epoch, k, prevTime, attack, net, optimizer

    k, prevTime = 0, 0

    net = None

    print("==> Building model..")

    if args.net == "cnn":
        net = nn.Sequential(
            nn.Conv2d(1, 32, 5, stride=1, padding=2),
            nn.ReLU(),
            nn.MaxPool2d(2),
            nn.Conv2d(32, 64, 5, stride=1, padding=2),
            nn.ReLU(),
            nn.MaxPool2d(2),
            Flatten(),
            nn.Linear(64 * 7 * 7, 1024),
            nn.ReLU(),
            nn.Linear(1024, 10),
        )

    elif args.net == "resnet50":
        net = ResNet50()

    # elif args.net=='densenet161':

    # net = DenseNet161()

    elif args.net == "wideresnet":
        net = WideResNet(
            depth=34, num_classes=100 if args.data == "cifar100" else 10
        )

    else:
        raise NotImplementedError()

    optimizer = None  # optim.Adam(net.parameters(), lr=1e-4)

    if args.resume:
        if args.resumefrom is None:
            model_dir = args.save_name

            all_models = [
                name
                for name in os.listdir(model_dir)
                if name.split(".")[-1] != "optim"
            ]

            assert len(all_models) >= 1, (
                "Error: no checkpoint file found in " + model_dir
            )

            args.resumefrom = os.path.join(
                model_dir,
                sorted(all_models, key=lambda x: int(x.split(".")[3]))[-1],
            )  # the biggest model

            print("args.resumefrom", args.resumefrom)

    if torch.cuda.is_available():
        net.cuda()

        net = torch.nn.DataParallel(
            net, device_ids=range(torch.cuda.device_count())
        )

        cudnn.benchmark = True

    if optimizer is None:
        optimizer = optim.Adam(net.parameters(), lr=1e-4)

    if args.sgd:
        optimizer = optim.SGD(
            net.parameters(), lr=args.lr, momentum=0.9, weight_decay=5e-4
        )

    epsilon, step_size = 0, 0

    if args.data == "mnist":
        epsilon, step_size = 0.3, 0.6

    elif args.data == "cifar10" or args.data == "cifar100":
        epsilon, step_size = 0.032, 0.064

    elif args.data == "svhn":
        epsilon, step_size = 0.047, 0.094

    attack = LinfPGDAttack(net, epsilon, 1, step_size, True, "xent")

    # Data

    print("==> Preparing data..")

    transform_train = transforms.Compose(
        [
            transforms.RandomCrop(32, padding=4),
            transforms.RandomHorizontalFlip(),
            transforms.ToTensor(),
        ]
    )

    transform_valid = transforms.Compose(
        [
            transforms.CenterCrop(32),
            transforms.ToTensor(),
        ]
    )

    trainset, validset = None, None

    valid_size = 0.02

    if ds_name == "cifar10":
        trainset = torchvision.datasets.CIFAR10(
            root="./data", train=True, download=True, transform=transform_train
        )

        validset = torchvision.datasets.CIFAR10(
            root="./data", train=True, download=True, transform=transform_valid
        )

    elif ds_name == "cifar100":
        trainset = torchvision.datasets.CIFAR100(
            root="./data", train=True, download=True, transform=transform_train
        )

        validset = torchvision.datasets.CIFAR100(
            root="./data", train=True, download=True, transform=transform_valid
        )

    else:
        raise NotImplementedError

    ##### get random valid indices

    valid_indices, train_indices = [], []

    randidxfn_valid = "./data/" + args.data + ".validk.randidx"

    randidxfn_train = "./data/" + args.data + ".traink.randidx"

    if os.path.exists(randidxfn_valid):  # load to **_randidx
        valid_indices = np.genfromtxt(randidxfn_valid).astype("int")

        train_indices = np.genfromtxt(randidxfn_train).astype("int")

    else:  # generate randidx, load to **_randidx and save to file
        trainloader = torch.utils.data.DataLoader(
            trainset, batch_size=args.batch_size, shuffle=False
        )

        train_labels = {}

        gid = 0

        for _, (_, targets) in enumerate(trainloader):
            for t in Variable(targets).data:
                t = int(t)

                if t in train_labels:
                    train_labels[t].append(gid)

                else:
                    train_labels[t] = [gid]

                gid += 1

        for tk, ti in train_labels.items():
            # np.random.seed(tk) # fix for reproducibility

            np.random.shuffle(ti)

            split_sz = int(valid_size * len(ti))

            valid_indices += ti[:split_sz]

            train_indices += ti[split_sz:]

        valid_indices, train_indices = (
            np.array(valid_indices),
            np.array(train_indices),
        )

        np.savetxt(randidxfn_valid, valid_indices, fmt="%.f")

        np.savetxt(randidxfn_train, train_indices, fmt="%.f")

    np.random.shuffle(train_indices)

    train_sampler = SubsetRandomSampler(train_indices)

    valid_sampler = SubsetRandomSampler(valid_indices)

    trainloader = torch.utils.data.DataLoader(
        trainset,
        batch_size=args.batch_size,
        sampler=train_sampler,
        num_workers=2,
    )

    validloader = torch.utils.data.DataLoader(
        validset,
        batch_size=args.batch_size,
        sampler=valid_sampler,
        num_workers=2,
    )

    epoch = 0

    while epoch < 1 + args.epochs:
        epoch = advtrain_CAT(epoch, trainloader, validloader, args)

        epoch += 1


# if __name__ == "__main__":

# main_currat("cifar10")



# (c) 2024 LAiSR-SK
# This code is licensed under the MIT license (see LICENSE.md).
from collections import namedtuple

import numpy as np
import torch
import torchvision

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

################################################################
## Components from https://github.com/davidcpage/cifar10-fast ##
################################################################

#####################
## data preprocessing
#####################

cifar10_mean = (
    0.4914,
    0.4822,
    0.4465,
)  # equals np.mean(train_set.train_data, axis=(0,1,2))/255
cifar10_std = (
    0.2471,
    0.2435,
    0.2616,
)  # equals np.std(train_set.train_data, axis=(0,1,2))/255


def normalise(x, mean=cifar10_mean, std=cifar10_std):
    x, mean, std = (np.array(a, np.float32) for a in (x, mean, std))
    x -= mean * 255
    x *= 1.0 / (255 * std)
    return x


def pad(x, border=4):
    return np.pad(
        x, [(0, 0), (border, border), (border, border), (0, 0)], mode="reflect"
    )


def transpose(x, source="NHWC", target="NCHW"):
    return x.transpose([source.index(d) for d in target])


#####################
## data augmentation
#####################


class Crop(namedtuple("Crop", ("h", "w"))):
    def __call__(self, x, x0, y0):
        return x[:, y0 : y0 + self.h, x0 : x0 + self.w]

    def options(self, x_shape):
        C, H, W = x_shape
        return {"x0": range(W + 1 - self.w), "y0": range(H + 1 - self.h)}

    def output_shape(self, x_shape):
        C, H, W = x_shape
        return (C, self.h, self.w)


class FlipLR(namedtuple("FlipLR", ())):
    def __call__(self, x, choice):
        return x[:, :, ::-1].copy() if choice else x

    def options(self, x_shape):
        return {"choice": [True, False]}


class Cutout(namedtuple("Cutout", ("h", "w"))):
    def __call__(self, x, x0, y0):
        x = x.copy()
        x[:, y0 : y0 + self.h, x0 : x0 + self.w].fill(0.0)
        return x

    def options(self, x_shape):
        C, H, W = x_shape
        return {"x0": range(W + 1 - self.w), "y0": range(H + 1 - self.h)}


class Transform:
    def __init__(self, dataset, transforms):
        self.dataset, self.transforms = dataset, transforms
        self.choices = None

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, index):
        data, labels = self.dataset[index]
        for choices, f in zip(self.choices, self.transforms):
            args = {k: v[index] for (k, v) in choices.items()}
            data = f(data, **args)
        return data, labels

    def set_random_choices(self):
        self.choices = []
        x_shape = self.dataset[0][0].shape
        N = len(self)
        for t in self.transforms:
            options = t.options(x_shape)
            x_shape = (
                t.output_shape(x_shape)
                if hasattr(t, "output_shape")
                else x_shape
            )
            self.choices.append(
                {k: np.random.choice(v, size=N) for (k, v) in options.items()}
            )


#####################
## dataset
#####################


def cifar10(root):
    train_set = torchvision.datasets.CIFAR10(
        root=root, train=True, download=True
    )
    test_set = torchvision.datasets.CIFAR10(
        root=root, train=False, download=True
    )
    return {
        "train": {"data": train_set.data, "labels": train_set.targets},
        "test": {"data": test_set.data, "labels": test_set.targets},
    }


#####################
## data loading
#####################


class Batches:
    def __init__(
        self,
        dataset,
        batch_size,
        shuffle,
        set_random_choices=False,
        num_workers=0,
        drop_last=False,
    ):
        self.dataset = dataset
        self.batch_size = batch_size
        self.set_random_choices = set_random_choices
        self.dataloader = torch.utils.data.DataLoader(
            dataset,
            batch_size=batch_size,
            num_workers=num_workers,
            pin_memory=True,
            shuffle=shuffle,
            drop_last=drop_last,
        )

    def __iter__(self):
        if self.set_random_choices:
            self.dataset.set_random_choices()
        return (
            {"input": x.to(device).half(), "target": y.to(device).long()}
            for (x, y) in self.dataloader
        )

    def __len__(self):
        return len(self.dataloader)



# (c) 2024 LAiSR-SK
# This code is licensed under the MIT license (see LICENSE.md).
import numpy as np
import torch
import torch.nn as nn
from torch.nn.functional import log_softmax, softmax


class FAT_EARLYSTOP:
    def __init__(self, model, device):
        self.model = model
        self.device = device

    def earlystop(
        self,
        data,
        target,
        step_size,
        epsilon,
        perturb_steps,
        tau,
        randominit_type,
        loss_fn,
        rand_init=True,
        omega=0,
    ):
        """
        The implematation of early-stopped PGD
        Following the Alg.1 in our FAT paper <https://arxiv.org/abs/2002.11242>
        :param step_size: the PGD step size
        :param epsilon: the perturbation bound
        :param perturb_steps: the maximum PGD step
        :param tau: the step controlling how early we should stop interations when wrong adv data is found
        :param randominit_type: To decide the type of random inirialization (random start for searching adv data)
        :param rand_init: To decide whether to initialize adversarial sample with random noise (random start for searching adv data)
        :param omega: random sample parameter for adv data generation (this is for escaping the local minimum.)
        :return: output_adv (friendly adversarial data) output_target (targets), output_natural (the corresponding natrual data), count (average backword propagations count)

        :note if you use grayscale images then you may need to adjust the shape output_adv and output_natural
        """
        model = self.model
        model.eval()

        K = perturb_steps
        count = 0
        output_target = []
        output_adv = []
        output_natural = []

        control = (torch.ones(len(target)) * tau).to(self.device)

        # Initialize the adversarial data with random noise
        if rand_init:
            if randominit_type == "normal_distribution_randominit":
                iter_adv = (
                    data.detach()
                    + 0.001 * torch.randn(data.shape).to(self.device).detach()
                )
                iter_adv = torch.clamp(iter_adv, 0.0, 1.0)
            if randominit_type == "uniform_randominit":
                iter_adv = data.detach() + torch.from_numpy(
                    np.random.uniform(-epsilon, epsilon, data.shape)
                ).float().to(self.device)
                iter_adv = torch.clamp(iter_adv, 0.0, 1.0)
        else:
            iter_adv = data.to(self.device).detach()

        iter_clean_data = data.to(self.device).detach()
        iter_target = target.to(self.device).detach()
        output_iter_clean_data = model(data)

        while K > 0:
            iter_adv.requires_grad_()
            output = model(iter_adv)
            pred = output.max(1, keepdim=True)[1]
            output_index = []
            iter_index = []

            # Calculate the indexes of adversarial data those still needs to be iterated
            for idx, (p, it) in enumerate(zip(pred, iter_target)):
                if p != it:
                    if control[idx] == 0:
                        output_index.append(idx)
                    else:
                        control[idx] -= 1
                        iter_index.append(idx)
                else:
                    iter_index.append(idx)

            # Add adversarial data those do not need any more iteration into set output_adv
            if len(output_index) != 0:
                if len(output_target) == 0:
                    # incorrect adv data should not keep iterated
                    output_adv = (
                        iter_adv[output_index]
                        .reshape(-1, 3, 32, 32)
                        .to(self.device)
                    )
                    output_natural = (
                        iter_clean_data[output_index]
                        .reshape(-1, 3, 32, 32)
                        .to(self.device)
                    )
                    output_target = (
                        iter_target[output_index].reshape(-1).to(self.device)
                    )
                else:
                    # incorrect adv data should not keep iterated
                    output_adv = torch.cat(
                        (
                            output_adv,
                            iter_adv[output_index]
                            .reshape(-1, 3, 32, 32)
                            .to(self.device),
                        ),
                        dim=0,
                    )
                    output_natural = torch.cat(
                        (
                            output_natural,
                            iter_clean_data[output_index]
                            .reshape(-1, 3, 32, 32)
                            .to(self.device),
                        ),
                        dim=0,
                    )
                    output_target = torch.cat(
                        (
                            output_target,
                            iter_target[output_index]
                            .reshape(-1)
                            .to(self.device),
                        ),
                        dim=0,
                    )

            # calculate gradient
            model.zero_grad()
            with torch.enable_grad():
                if loss_fn == "cent":
                    loss_adv = nn.CrossEntropyLoss(reduction="mean")(
                        output, iter_target
                    )
                if loss_fn == "kl":
                    criterion_kl = nn.KLDivLoss(size_average=False).to(
                        self.device
                    )
                    loss_adv = criterion_kl(
                        log_softmax(output, dim=1),
                        softmax(output_iter_clean_data, dim=1),
                    )
            loss_adv.backward(retain_graph=True)
            grad = iter_adv.grad

            # update iter adv
            if len(iter_index) != 0:
                control = control[iter_index]
                iter_adv = iter_adv[iter_index]
                iter_clean_data = iter_clean_data[iter_index]
                iter_target = iter_target[iter_index]
                output_iter_clean_data = output_iter_clean_data[iter_index]
                grad = grad[iter_index]
                eta = step_size * grad.sign()

                iter_adv = (
                    iter_adv.detach()
                    + eta
                    + omega
                    * torch.randn(iter_adv.shape).detach().to(self.device)
                )
                iter_adv = torch.min(
                    torch.max(iter_adv, iter_clean_data - epsilon),
                    iter_clean_data + epsilon,
                )
                iter_adv = torch.clamp(iter_adv, 0, 1)
                count += len(iter_target)
            else:
                output_adv = output_adv.detach()
                return output_adv, output_target, output_natural, count
            K = K - 1

        if len(output_target) == 0:
            output_target = iter_target.reshape(-1).squeeze().to(self.device)
            output_adv = iter_adv.reshape(-1, 3, 32, 32).to(self.device)
            output_natural = iter_clean_data.reshape(-1, 3, 32, 32).to(
                self.device
            )
        else:
            output_adv = torch.cat(
                (output_adv, iter_adv.reshape(-1, 3, 32, 32)), dim=0
            ).to(self.device)
            output_target = (
                torch.cat((output_target, iter_target.reshape(-1)), dim=0)
                .squeeze()
                .to(self.device)
            )
            output_natural = torch.cat(
                (
                    output_natural,
                    iter_clean_data.reshape(-1, 3, 32, 32).to(self.device),
                ),
                dim=0,
            ).to(self.device)
        output_adv = output_adv.detach()
        return output_adv, output_target, output_natural, count



import numpy as np
import torch

from ..attack import Attack
from .deepfool import DeepFool


class SparseFool(Attack):
    r"""
    Attack in the paper 'SparseFool: a few pixels make a big difference'
    [https://arxiv.org/abs/1811.02248]

    Modified from "https://github.com/LTS4/SparseFool/"

    Distance Measure : L0

    Arguments:
        model (nn.Module): model to attack.
        steps (int): number of steps. (Default: 10)
        lam (float): parameter for scaling DeepFool noise. (Default: 3)
        overshoot (float): parameter for enhancing the noise. (Default: 0.02)

    Shape:
        - images: :math:`(N, C, H, W)` where `N = number of batches`, `C = number of channels`,        `H = height` and `W = width`. It must have a range [0, 1].
        - labels: :math:`(N)` where each value :math:`y_i` is :math:`0 \leq y_i \leq` `number of labels`.
        - output: :math:`(N, C, H, W)`.

    Examples::
        >>> attack = torchattacks.SparseFool(model, steps=10, lam=3, overshoot=0.02)
        >>> adv_images = attack(images, labels)

    """

    def __init__(self, model, steps=10, lam=3, overshoot=0.02):
        super().__init__("SparseFool", model)
        self.steps = steps
        self.lam = lam
        self.overshoot = overshoot
        self.deepfool = DeepFool(model)
        self.supported_mode = ["default"]

    def forward(self, images, labels):
        r"""
        Overridden.
        """

        images = images.clone().detach().to(self.device)
        labels = labels.clone().detach().to(self.device)

        batch_size = len(images)
        correct = torch.tensor([True] * batch_size)
        curr_steps = 0

        adv_images = []
        for idx in range(batch_size):
            image = images[idx : idx + 1].clone().detach()
            adv_images.append(image)

        while (True in correct) and (curr_steps < self.steps):
            for idx in range(batch_size):
                image = images[idx : idx + 1]
                label = labels[idx : idx + 1]
                adv_image = adv_images[idx]

                fs = self.get_logits(adv_image)[0]
                _, pre = torch.max(fs, dim=0)
                if pre != label:
                    correct[idx] = False
                    continue

                adv_image, target_label = (
                    self.deepfool.forward_return_target_labels(
                        adv_image, label
                    )
                )
                adv_image = image + self.lam * (adv_image - image)

                adv_image.requires_grad = True
                fs = self.get_logits(adv_image)[0]
                _, pre = torch.max(fs, dim=0)

                if pre == label:
                    pre = target_label

                cost = fs[pre] - fs[label]
                grad = torch.autograd.grad(
                    cost, adv_image, retain_graph=False, create_graph=False
                )[0]
                grad = grad / grad.norm()

                adv_image = self._linear_solver(image, grad, adv_image)
                adv_image = image + (1 + self.overshoot) * (adv_image - image)
                adv_images[idx] = torch.clamp(adv_image, min=0, max=1).detach()

            curr_steps += 1

        adv_images = torch.cat(adv_images).detach()

        return adv_images

    def _linear_solver(self, x_0, coord_vec, boundary_point):
        input_shape = x_0.size()

        plane_normal = coord_vec.clone().detach().view(-1)
        plane_point = boundary_point.clone().detach().view(-1)

        x_i = x_0.clone().detach()

        f_k = torch.dot(plane_normal, x_0.view(-1) - plane_point)
        sign_true = f_k.sign().item()

        beta = 0.001 * sign_true
        current_sign = sign_true

        while current_sign == sign_true and coord_vec.nonzero().size()[0] > 0:
            f_k = torch.dot(plane_normal, x_i.view(-1) - plane_point) + beta

            pert = f_k.abs() / coord_vec.abs().max()

            mask = torch.zeros_like(coord_vec)
            mask[
                np.unravel_index(
                    torch.argmax(coord_vec.abs()).cpu(), input_shape
                )
            ] = 1.0  # nopep8

            r_i = torch.clamp(pert, min=1e-4) * mask * coord_vec.sign()

            x_i = x_i + r_i
            x_i = torch.clamp(x_i, min=0, max=1)

            f_k = torch.dot(plane_normal, x_i.view(-1) - plane_point)
            current_sign = f_k.sign().item()

            coord_vec[r_i != 0] = 0

        return x_i



import numpy as np

import torch
import torch.nn as nn

from ..attack import Attack


class UPGD(Attack):
    r"""
    Ultimate PGD that supports various options of gradient-based adversarial attacks.

    Distance Measure : Linf

    Arguments:
        model (nn.Module): model to attack.
        eps (float): maximum perturbation. (Default: 8/255)
        alpha (float): step size. (Default: 2/255)
        steps (int): number of steps. (Default: 10)
        random_start (bool): using random initialization of delta. (Default: False)
        loss (str): loss function. ['ce', 'margin', 'dlr'] (Default: 'ce')
        decay (float): momentum factor. (Default: 1.0)
        eot_iter (int) : number of models to estimate the mean gradient. (Default: 1)

    Shape:
        - images: :math:`(N, C, H, W)` where `N = number of batches`, `C = number of channels`,        `H = height` and `W = width`. It must have a range [0, 1].
        - labels: :math:`(N)` where each value :math:`y_i` is :math:`0 \leq y_i \leq` `number of labels`.
        - output: :math:`(N, C, H, W)`.

    Examples::
        >>> attack = torchattacks.UPGD(model, eps=8/255, alpha=2/255, steps=10, random_start=False)
        >>> adv_images = attack(images, labels)

    """

    def __init__(
        self,
        model,
        eps=8 / 255,
        alpha=2 / 255,
        steps=10,
        random_start=False,
        loss="ce",
        decay=1.0,
        eot_iter=1,
    ):
        super().__init__("UPGD", model)
        self.eps = eps
        self.alpha = alpha
        self.steps = steps
        self.random_start = random_start
        self.loss = loss
        self.decay = decay
        self.eot_iter = eot_iter
        self.supported_mode = ["default", "targeted"]

    def forward(self, images, labels):
        r"""
        Overridden.
        """

        images = images.clone().detach().to(self.device)
        labels = labels.clone().detach().to(self.device)

        if self.targeted:
            target_labels = self.get_target_label(images, labels)

        momentum = torch.zeros_like(images).detach().to(self.device)

        adv_images = images.clone().detach()

        if self.random_start:
            # Starting at a uniformly random point
            adv_images = adv_images + torch.empty_like(adv_images).uniform_(
                -self.eps, self.eps
            )
            adv_images = torch.clamp(adv_images, min=0, max=1).detach()

        for _ in range(self.steps):
            grad = torch.zeros_like(adv_images)
            adv_images.requires_grad = True

            for j in range(self.eot_iter):
                # Calculate loss
                if self.targeted:
                    cost = self.get_loss(adv_images, labels, target_labels)
                else:
                    cost = self.get_loss(adv_images, labels)

                grad += (
                    torch.autograd.grad(
                        cost, adv_images, retain_graph=False, create_graph=False
                    )[0]
                    / self.eot_iter
                )

            # Update adversarial images
            grad = grad / torch.mean(torch.abs(grad), dim=(1, 2, 3), keepdim=True)
            grad = grad + momentum * self.decay
            momentum = grad

            adv_images = adv_images.detach() + self.alpha * grad.sign()
            delta = torch.clamp(adv_images - images, min=-self.eps, max=self.eps)
            adv_images = torch.clamp(images + delta, min=0, max=1).detach()

        return adv_images

    def get_loss(self, images, labels, target_labels=None):
        if isinstance(self.loss, str):
            if self.loss == "ce":
                return self.ce_loss(images, labels, target_labels)
            elif self.loss == "dlr":
                return self.dlr_loss(images, labels, target_labels)
            elif self.loss == "margin":
                return self.margin_loss(images, labels, target_labels)
            else:
                raise ValueError(self.loss + " is not valid.")
        else:
            return self.loss(images, labels, target_labels)

    def ce_loss(self, images, labels, target_labels):
        loss = nn.CrossEntropyLoss()
        outputs = self.get_logits(images)

        if self.targeted:
            cost = -loss(outputs, target_labels)
        else:
            cost = loss(outputs, labels)
        return cost

    def dlr_loss(self, images, labels, target_labels):
        outputs = self.get_logits(images)
        outputs_sorted, ind_sorted = outputs.sort(dim=1)
        if self.targeted:
            cost = -(
                outputs[np.arange(outputs.shape[0]), labels]
                - outputs[np.arange(outputs.shape[0]), target_labels]
            ) / (
                outputs_sorted[:, -1]
                - 0.5 * outputs_sorted[:, -3]
                - 0.5 * outputs_sorted[:, -4]
                + 1e-12
            )
        else:
            ind = (ind_sorted[:, -1] == labels).float()
            cost = -(
                outputs[np.arange(outputs.shape[0]), labels]
                - outputs_sorted[:, -2] * ind
                - outputs_sorted[:, -1] * (1.0 - ind)
            ) / (outputs_sorted[:, -1] - outputs_sorted[:, -3] + 1e-12)
        return cost.sum()

    # f-function in the paper
    def margin_loss(self, images, labels, target_labels):
        outputs = self.get_logits(images)
        if self.targeted:
            # one_hot_labels = torch.eye(len(outputs[0]))[target_labels].to(self.device)
            one_hot_labels = torch.eye(outputs.shape[1]).to(self.device)[
                target_labels
            ]  # nopep8
            i, _ = torch.max((1 - one_hot_labels) * outputs, dim=1)
            j = torch.masked_select(outputs, one_hot_labels.bool())
            cost = -torch.clamp((i - j), min=0)  # -self.kappa=0
        else:
            one_hot_labels = torch.eye(len(outputs[0]))[labels].to(self.device)
            i, _ = torch.max((1 - one_hot_labels) * outputs, dim=1)
            j = torch.masked_select(outputs, one_hot_labels.bool())
            cost = -torch.clamp((j - i), min=0)  # -self.kappa=0
        return cost.sum()



import torch
import numpy as np

from ..attack import Attack


class JSMA(Attack):
    r"""
    Jacobian Saliency Map Attack in the paper 'The Limitations of Deep Learning in Adversarial Settings'
    [https://arxiv.org/abs/1511.07528v1]

    Distance Measure : L0

    Arguments:
        model (nn.Module): model to attack.
        theta (float): perturb length, range is either [theta, 0], [0, theta]. (Default: 1.0)
        gamma (float): highest percentage of pixels can be modified. (Default: 0.1)

    Shape:
        - images: :math:`(N, C, H, W)` where `N = number of batches`, `C = number of channels`,        `H = height` and `W = width`. It must have a range [0, 1].
        - labels: :math:`(N)` where each value :math:`y_i` is :math:`0 \leq y_i \leq` `number of labels`.
        - output: :math:`(N, C, H, W)`.

    Examples::
        >>> attack = torchattacks.JSMA(model, theta=1.0, gamma=0.1)
        >>> adv_images = attack(images, labels)

    """

    def __init__(self, model, theta=1.0, gamma=0.1):
        super().__init__("JSMA", model)
        self.theta = theta
        self.gamma = gamma
        self.supported_mode = ["default", "targeted"]

    def forward(self, images, labels):
        r"""
        Overridden.
        """

        images = images.clone().detach().to(self.device)
        labels = labels.clone().detach().to(self.device)

        if self.targeted:
            target_labels = self.get_target_label(images, labels)
        else:
            # Because the JSMA algorithm does not use any loss function,
            # it cannot perform untargeted attacks indeed
            # (we have no control over the convergence of the attack to a data point that is NOT equal to the original class),
            # so we make the default setting of the target label is right circular shift
            # to make attack work if user didn't set target label.
            target_labels = (labels + 1) % 10

        adv_images = None
        for im, tl in zip(images, target_labels):
            # Since the attack uses the Jacobian-matrix,
            # if we input a large number of images directly into it,
            # the processing will be very complicated,
            # here, in order to simplify the processing,
            # we only process one image at a time.
            # Shape of MNIST is [-1, 1, 28, 28],
            # and shape of CIFAR10 is [-1, 3, 32, 32].
            pert_image = self.perturbation_single(
                torch.unsqueeze(im, 0), torch.unsqueeze(tl, 0)
            )
            try:
                adv_images = torch.cat((adv_images, pert_image), 0)
            except Exception:
                adv_images = pert_image

        adv_images = torch.clamp(adv_images, min=0, max=1)
        return adv_images

    def compute_jacobian(self, image):
        var_image = image.clone().detach()
        var_image.requires_grad = True
        output = self.get_logits(var_image)

        num_features = int(np.prod(var_image.shape[1:]))
        jacobian = torch.zeros([output.shape[1], num_features])
        for i in range(output.shape[1]):
            if var_image.grad is not None:
                var_image.grad.zero_()
            output[0][i].backward(retain_graph=True)
            # Copy the derivative to the target place
            jacobian[i] = (
                var_image.grad.squeeze().view(-1, num_features).clone()
            )  # nopep8

        return jacobian.to(self.device)

    @torch.no_grad()
    def saliency_map(
        self, jacobian, target_label, increasing, search_space, nb_features
    ):
        # The search domain
        domain = torch.eq(search_space, 1).float()
        # The sum of all features' derivative with respect to each class
        all_sum = torch.sum(jacobian, dim=0, keepdim=True)
        # The forward derivative of the target class
        target_grad = jacobian[target_label]
        # The sum of forward derivative of other classes
        others_grad = all_sum - target_grad

        # This list blanks out those that are not in the search domain
        if increasing:
            increase_coef = 2 * (torch.eq(domain, 0)).float().to(self.device)
        else:
            increase_coef = -1 * 2 * (torch.eq(domain, 0)).float().to(self.device)
        increase_coef = increase_coef.view(-1, nb_features)

        # Calculate sum of target forward derivative of any 2 features.
        target_tmp = target_grad.clone()
        target_tmp -= increase_coef * torch.max(torch.abs(target_grad))
        # PyTorch will automatically extend the dimensions
        alpha = target_tmp.view(-1, 1, nb_features) + target_tmp.view(
            -1, nb_features, 1
        )
        # Calculate sum of other forward derivative of any 2 features.
        others_tmp = others_grad.clone()
        others_tmp += increase_coef * torch.max(torch.abs(others_grad))
        beta = others_tmp.view(-1, 1, nb_features) + others_tmp.view(-1, nb_features, 1)

        # Zero out the situation where a feature sums with itself
        tmp = np.ones((nb_features, nb_features), int)
        np.fill_diagonal(tmp, 0)
        zero_diagonal = torch.from_numpy(tmp).byte().to(self.device)

        # According to the definition of saliency map in the paper (formulas 8 and 9),
        # those elements in the saliency map that doesn't satisfy the requirement will be blanked out.
        if increasing:
            mask1 = torch.gt(alpha, 0.0)
            mask2 = torch.lt(beta, 0.0)
        else:
            mask1 = torch.lt(alpha, 0.0)
            mask2 = torch.gt(beta, 0.0)

        # Apply the mask to the saliency map
        mask = torch.mul(torch.mul(mask1, mask2), zero_diagonal.view_as(mask1))
        # Do the multiplication according to formula 10 in the paper
        saliency_map = torch.mul(torch.mul(alpha, torch.abs(beta)), mask.float())
        # Get the most significant two pixels
        max_idx = torch.argmax(saliency_map.view(-1, nb_features * nb_features), dim=1)
        # p = max_idx // nb_features
        p = torch.div(max_idx, nb_features, rounding_mode="floor")
        # q = max_idx % nb_features
        q = max_idx - p * nb_features
        return p, q

    def perturbation_single(self, image, target_label):
        """
        image: only one element
        label: only one element
        """
        var_image = image
        var_label = target_label
        var_image = var_image.to(self.device)
        var_label = var_label.to(self.device)

        if self.theta > 0:
            increasing = True
        else:
            increasing = False

        num_features = int(np.prod(var_image.shape[1:]))
        shape = var_image.shape

        # Perturb two pixels in one iteration, thus max_iters is divided by 2
        max_iters = int(np.ceil(num_features * self.gamma / 2.0))

        # Masked search domain, if the pixel has already reached the top or bottom, we don't bother to modify it
        if increasing:
            search_domain = torch.lt(var_image, 0.99)
        else:
            search_domain = torch.gt(var_image, 0.01)

        search_domain = search_domain.view(num_features)
        output = self.get_logits(var_image)
        current_pred = torch.argmax(output.data, 1)

        iter = 0
        while (
            (iter < max_iters)
            and (current_pred != target_label)
            and (search_domain.sum() != 0)
        ):
            # Calculate Jacobian matrix of forward derivative
            jacobian = self.compute_jacobian(var_image)
            # Get the saliency map and calculate the two pixels that have the greatest influence
            p1, p2 = self.saliency_map(
                jacobian, var_label, increasing, search_domain, num_features
            )
            # Apply modifications
            # var_sample_flatten = var_image.view(-1, num_features).clone().detach_()
            var_sample_flatten = var_image.view(-1, num_features)
            var_sample_flatten[0, p1] += self.theta
            var_sample_flatten[0, p2] += self.theta

            new_image = torch.clamp(var_sample_flatten, min=0.0, max=1.0)
            new_image = new_image.view(shape)
            search_domain[p1] = 0
            search_domain[p2] = 0
            # var_image = new_image.clone().detach().to(self.device)
            var_image = new_image.to(self.device)

            output = self.get_logits(var_image)
            current_pred = torch.argmax(output.data, 1)
            iter += 1

        adv_image = var_image
        return adv_image



import torch
import torch.nn as nn

from ..attack import Attack


class EOTPGD(Attack):
    r"""
    Comment on "Adv-BNN: Improved Adversarial Defense through Robust Bayesian Neural Network"
    [https://arxiv.org/abs/1907.00895]

    Distance Measure : Linf

    Arguments:
        model (nn.Module): model to attack.
        eps (float): maximum perturbation. (Default: 8/255)
        alpha (float): step size. (Default: 2/255)
        steps (int): number of steps. (Default: 10)
        eot_iter (int) : number of models to estimate the mean gradient. (Default: 2)

    Shape:
        - images: :math:`(N, C, H, W)` where `N = number of batches`, `C = number of channels`, `H = height` and `W = width`. It must have a range [0, 1].
        - labels: :math:`(N)` where each value :math:`y_i` is :math:`0 \leq y_i \leq` `number of labels`.
        - output: :math:`(N, C, H, W)`.

    Examples::
        >>> attack = torchattacks.EOTPGD(model, eps=8/255, alpha=2/255, steps=10, eot_iter=2)
        >>> adv_images = attack(images, labels)

    """

    def __init__(
        self,
        model,
        eps=8 / 255,
        alpha=2 / 255,
        steps=10,
        eot_iter=2,
        random_start=True,
    ):
        super().__init__("EOTPGD", model)
        self.eps = eps
        self.alpha = alpha
        self.steps = steps
        self.eot_iter = eot_iter
        self.random_start = random_start
        self.supported_mode = ["default", "targeted"]

    def forward(self, images, labels):
        r"""
        Overridden.
        """

        images = images.clone().detach().to(self.device)
        labels = labels.clone().detach().to(self.device)

        if self.targeted:
            target_labels = self.get_target_label(images, labels)

        loss = nn.CrossEntropyLoss()

        adv_images = images.clone().detach()

        if self.random_start:
            # Starting at a uniformly random point
            adv_images = adv_images + torch.empty_like(adv_images).uniform_(
                -self.eps, self.eps
            )  # nopep8
            adv_images = torch.clamp(adv_images, min=0, max=1).detach()

        for _ in range(self.steps):
            grad = torch.zeros_like(adv_images)
            adv_images.requires_grad = True

            for j in range(self.eot_iter):
                outputs = self.get_logits(adv_images)

                # Calculate loss
                if self.targeted:
                    cost = -loss(outputs, target_labels)
                else:
                    cost = loss(outputs, labels)

                # Update adversarial images
                grad += torch.autograd.grad(
                    cost, adv_images, retain_graph=False, create_graph=False
                )[0]

            # (grad/self.eot_iter).sign() == grad.sign()
            adv_images = adv_images.detach() + self.alpha * grad.sign()
            delta = torch.clamp(
                adv_images - images, min=-self.eps, max=self.eps
            )
            adv_images = torch.clamp(images + delta, min=0, max=1).detach()

        return adv_images



# (c) 2024 LAiSR-SK

# This code is licensed under the MIT license (see LICENSE.md).

import torch

from airtk.util import replicate_input


class Attack:
    """

    Abstract base class for all attack classes.

    Arguments:

        predict (nn.Module): forward pass function.

        loss_fn (nn.Module): loss function.

        clip_min (float): mininum value per input dimension.

        clip_max (float): maximum value per input dimension.

    """

    def __init__(self, predict, loss_fn, clip_min, clip_max):
        self.predict = predict

        self.loss_fn = loss_fn

        self.clip_min = clip_min

        self.clip_max = clip_max

    def perturb(self, x, **kwargs):
        """

        Virtual method for generating the adversarial examples.

        Arguments:

            x (torch.Tensor): the model's input tensor.

            **kwargs: optional parameters used by child classes.

        Returns:

            adversarial examples.

        """

        error = "Sub-classes must implement perturb."

        raise NotImplementedError(error)

    def __call__(self, *args, **kwargs):
        return self.perturb(*args, **kwargs)


class LabelMixin:
    def _get_predicted_label(self, x):
        """

        Compute predicted labels given x. Used to prevent label leaking during adversarial training.

        Arguments:

            x (torch.Tensor): the model's input tensor.

        Returns:

            torch.Tensor containing predicted labels.

        """

        with torch.no_grad():
            outputs = self.predict(x)

        _, y = torch.max(outputs, dim=1)

        return y

    def _verify_and_process_inputs(self, x, y):
        if self.targeted:
            assert y is not None

        if not self.targeted:
            if y is None:
                y = self._get_predicted_label(x)

        x = replicate_input(x)

        y = replicate_input(y)

        return x, y



from itertools import chain

import numpy as np
import torch
from torch.nn.functional import softmax

from ..attack import Attack


class Pixle(Attack):
    r"""
    Pixle: a fast and effective black-box attack based on rearranging pixels'
    [https://arxiv.org/abs/2202.02236]

    Distance Measure : L0

    Arguments:
        model (nn.Module): model to attack.
        x_dimensions (int or float, or a tuple containing a combination of those): size of the sampled patch along ther x side for each iteration. The integers are considered as fixed number of size,
        while the float as parcentage of the size. A tuple is used to specify both under and upper bound of the size. (Default: (2, 10))
        y_dimensions (int or float, or a tuple containing a combination of those): size of the sampled patch along ther y side for each iteration. The integers are considered as fixed number of size,
        while the float as parcentage of the size. A tuple is used to specify both under and upper bound of the size. (Default: (2, 10))
        pixel_mapping (str): the type of mapping used to move the pixels. Can be: 'random', 'similarity', 'similarity_random', 'distance', 'distance_random' (Default: random)
        restarts (int): the number of restarts that the algortihm performs. (Default: 20)
        max_iterations (int): number of iterations to perform for each restart. (Default: 10)
        update_each_iteration (bool): if the attacked images must be modified after each iteration (True) or after each restart (False).  (Default: False)
    Shape:
        - images: :math:`(N, C, H, W)` where `N = number of batches`, `C = number of channels`,        `H = height` and `W = width`. It must have a range [0, 1].
        - labels: :math:`(N)` where each value :math:`y_i` is :math:`0 \leq y_i \leq` `number of labels`.
        - output: :math:`(N, C, H, W)`.

    Examples::
        >>> attack = torchattacks.Pixle(model, x_dimensions=(0.1, 0.2), restarts=10, max_iterations=50)
        >>> adv_images = attack(images, labels)
    """

    def __init__(
        self,
        model,
        x_dimensions=(2, 10),
        y_dimensions=(2, 10),
        pixel_mapping="random",
        restarts=20,
        max_iterations=10,
        update_each_iteration=False,
    ):
        super().__init__("Pixle", model)

        if restarts < 0 or not isinstance(restarts, int):
            raise ValueError(
                "restarts must be and integer >= 0 " f"({restarts})"
            )

        self.update_each_iteration = update_each_iteration
        self.max_patches = max_iterations

        self.restarts = restarts
        self.pixel_mapping = pixel_mapping.lower()

        if self.pixel_mapping not in [
            "random",
            "similarity",
            "similarity_random",
            "distance",
            "distance_random",
        ]:
            raise ValueError(
                "pixel_mapping must be one of [random, similarity,"
                "similarity_random, distance, distance_random]"
                f" ({self.pixel_mapping})"
            )

        if isinstance(y_dimensions, (int, float)):
            y_dimensions = [y_dimensions, y_dimensions]

        if isinstance(x_dimensions, (int, float)):
            x_dimensions = [x_dimensions, x_dimensions]

        if not all(
            [
                (isinstance(d, (int)) and d > 0)
                or (isinstance(d, float) and 0 <= d <= 1)
                for d in chain(y_dimensions, x_dimensions)
            ]
        ):
            raise ValueError(
                "dimensions of first patch must contains integers"
                " or floats in [0, 1]"
                f" ({y_dimensions})"
            )

        self.p1_x_dimensions = x_dimensions
        self.p1_y_dimensions = y_dimensions

        self.supported_mode = ["default", "targeted"]

    def forward(self, images, labels):
        if not self.update_each_iteration:
            adv_images = self.restart_forward(images, labels)
            return adv_images
        else:
            adv_images = self.iterative_forward(images, labels)
            return adv_images

    def restart_forward(self, images, labels):
        if len(images.shape) == 3:
            images = images.unsqueeze(0)

        if self.targeted:
            labels = self.get_target_label(images, labels)

        x_bounds = tuple(
            [
                max(1, d if isinstance(d, int) else round(images.size(3) * d))
                for d in self.p1_x_dimensions
            ]
        )

        y_bounds = tuple(
            [
                max(1, d if isinstance(d, int) else round(images.size(2) * d))
                for d in self.p1_y_dimensions
            ]
        )

        adv_images = []

        images = images.clone().detach().to(self.device)
        labels = labels.clone().detach().to(self.device)

        bs, _, _, _ = images.shape

        for idx in range(bs):
            image, label = images[idx : idx + 1], labels[idx : idx + 1]

            best_image = image.clone()
            pert_image = image.clone()

            loss, callback = self._get_fun(
                image, label, target_attack=self.targeted
            )
            best_solution = None

            best_p = loss(solution=image, solution_as_perturbed=True)
            image_probs = [best_p]

            it = 0

            for r in range(self.restarts):
                stop = False

                for it in range(self.max_patches):
                    (x, y), (x_offset, y_offset) = self.get_patch_coordinates(
                        image=image, x_bounds=x_bounds, y_bounds=y_bounds
                    )

                    destinations = self.get_pixel_mapping(
                        image,
                        x,
                        x_offset,
                        y,
                        y_offset,
                        destination_image=best_image,
                    )

                    solution = [x, y, x_offset, y_offset] + destinations

                    pert_image = self._perturb(
                        source=image, destination=best_image, solution=solution
                    )

                    p = loss(solution=pert_image, solution_as_perturbed=True)

                    if p < best_p:
                        best_p = p
                        best_solution = pert_image

                    image_probs.append(best_p)

                    if callback(pert_image, None, True):
                        best_solution = pert_image
                        stop = True
                        break

                if best_solution is None:
                    best_image = pert_image
                else:
                    best_image = best_solution

                if stop:
                    break

            adv_images.append(best_image)

        adv_images = torch.cat(adv_images)

        return adv_images

    def iterative_forward(self, images, labels):
        assert len(images.shape) == 3 or (
            len(images.shape) == 4 and images.size(0) == 1
        )

        if len(images.shape) == 3:
            images = images.unsqueeze(0)

        if self.targeted:
            labels = self.get_target_label(images, labels)

        x_bounds = tuple(
            [
                max(1, d if isinstance(d, int) else round(images.size(3) * d))
                for d in self.p1_x_dimensions
            ]
        )

        y_bounds = tuple(
            [
                max(1, d if isinstance(d, int) else round(images.size(2) * d))
                for d in self.p1_y_dimensions
            ]
        )

        adv_images = []

        images = images.clone().detach().to(self.device)
        labels = labels.clone().detach().to(self.device)

        bs, _, _, _ = images.shape

        for idx in range(bs):
            image, label = images[idx : idx + 1], labels[idx : idx + 1]

            best_image = image.clone()

            loss, callback = self._get_fun(
                image, label, target_attack=self.targeted
            )

            best_p = loss(solution=image, solution_as_perturbed=True)
            image_probs = [best_p]

            for it in range(self.max_patches):
                (x, y), (x_offset, y_offset) = self.get_patch_coordinates(
                    image=image, x_bounds=x_bounds, y_bounds=y_bounds
                )

                destinations = self.get_pixel_mapping(
                    image,
                    x,
                    x_offset,
                    y,
                    y_offset,
                    destination_image=best_image,
                )

                solution = [x, y, x_offset, y_offset] + destinations

                pert_image = self._perturb(
                    source=image, destination=best_image, solution=solution
                )

                p = loss(solution=pert_image, solution_as_perturbed=True)

                if p < best_p:
                    best_p = p
                    best_image = pert_image

                image_probs.append(best_p)

                if callback(pert_image, None, True):
                    best_image = pert_image
                    break

            adv_images.append(best_image)

        adv_images = torch.cat(adv_images)

        return adv_images

    def _get_prob(self, image):
        out = self.get_logits(image.to(self.device))
        prob = softmax(out, dim=1)
        return prob.detach().cpu().numpy()

    def loss(self, img, label, target_attack=False):
        p = self._get_prob(img)
        p = p[np.arange(len(p)), label]

        if target_attack:
            p = 1 - p

        return p.sum()

    def get_patch_coordinates(self, image, x_bounds, y_bounds):
        c, h, w = image.shape[1:]

        x, y = np.random.uniform(0, 1, 2)

        x_offset = np.random.randint(x_bounds[0], x_bounds[1] + 1)

        y_offset = np.random.randint(y_bounds[0], y_bounds[1] + 1)

        x, y = int(x * (w - 1)), int(y * (h - 1))

        if x + x_offset > w:
            x_offset = w - x

        if y + y_offset > h:
            y_offset = h - y

        return (x, y), (x_offset, y_offset)

    def get_pixel_mapping(
        self, source_image, x, x_offset, y, y_offset, destination_image=None
    ):
        if destination_image is None:
            destination_image = source_image

        destinations = []
        c, h, w = source_image.shape[1:]
        source_image = source_image[0]

        if self.pixel_mapping == "random":
            for i in range(x_offset):
                for j in range(y_offset):
                    dx, dy = np.random.uniform(0, 1, 2)
                    dx, dy = int(dx * (w - 1)), int(dy * (h - 1))
                    destinations.append([dx, dy])
        else:
            for i in np.arange(y, y + y_offset):
                for j in np.arange(x, x + x_offset):
                    pixel = source_image[:, i : i + 1, j : j + 1]
                    diff = destination_image - pixel
                    diff = diff[0].abs().mean(0).view(-1)

                    if "similarity" in self.pixel_mapping:
                        diff = 1 / (1 + diff)
                        diff[diff == 1] = 0

                    probs = torch.softmax(diff, 0).cpu().numpy()

                    indexes = np.arange(len(diff))

                    pair = None

                    linear_iter = iter(
                        sorted(
                            zip(indexes, probs),
                            key=lambda pit: pit[1],
                            reverse=True,
                        )
                    )

                    while True:
                        if "random" in self.pixel_mapping:
                            index = np.random.choice(indexes, p=probs)
                        else:
                            index = next(linear_iter)[0]

                        _y, _x = np.unravel_index(index, (h, w))

                        if _y == i and _x == j:
                            continue

                        pair = (_x, _y)
                        break

                    destinations.append(pair)

        return destinations

    def _get_fun(self, img, label, target_attack=False):
        img = img.to(self.device)

        if isinstance(label, torch.Tensor):
            label = label.cpu().numpy()

        @torch.no_grad()
        def func(
            solution, destination=None, solution_as_perturbed=False, **kwargs
        ):
            if not solution_as_perturbed:
                pert_image = self._perturb(
                    source=img, destination=destination, solution=solution
                )
            else:
                pert_image = solution

            p = self._get_prob(pert_image)
            p = p[np.arange(len(p)), label]

            if target_attack:
                p = 1 - p

            return p.sum()

        @torch.no_grad()
        def callback(
            solution, destination=None, solution_as_perturbed=False, **kwargs
        ):
            if not solution_as_perturbed:
                pert_image = self._perturb(
                    source=img, destination=destination, solution=solution
                )
            else:
                pert_image = solution

            p = self._get_prob(pert_image)[0]
            mx = np.argmax(p)

            if target_attack:
                return mx == label
            else:
                return mx != label

        return func, callback

    def _perturb(self, source, solution, destination=None):
        if destination is None:
            destination = source

        c, h, w = source.shape[1:]

        x, y, xl, yl = solution[:4]
        destinations = solution[4:]

        source_pixels = np.ix_(
            range(c), np.arange(y, y + yl), np.arange(x, x + xl)
        )

        indexes = torch.tensor(destinations)
        destination = destination.clone().detach().to(self.device)

        s = source[0][source_pixels].view(c, -1)

        destination[0, :, indexes[:, 0], indexes[:, 1]] = s

        return destination



# (c) 2024 LAiSR-SK
# This code is licensed under the MIT license (see LICENSE.md).
import torch
from torch import nn
from torch.autograd import Variable

device = "cuda" if torch.cuda.is_available() else "cpu"


def zero_gradients(x):
    """Manually zeroes the gradients when pytorch is unable"""

    if x.grad is not None:
        x.grad.zero_()


def sinkhorn_loss_joint_IPOT(
    alpha, beta, x_feature, y_feature, x_label, y_label, epsilon, m, n
):
    C_fea = get_cost_matrix(x_feature, y_feature)
    C = C_fea
    T = sinkhorn(C, 0.01, 100)
    # T = IPOT(C, 1)
    cost_ot = torch.sum(T * C)
    return cost_ot


def sinkhorn(C, epsilon, niter=50, device="cuda"):
    m = C.size(0)
    n = C.size(1)
    mu = Variable(
        1.0 / m * torch.FloatTensor(m).fill_(1).to("cuda"), requires_grad=False
    )
    nu = Variable(
        1.0 / n * torch.FloatTensor(n).fill_(1).to("cuda"), requires_grad=False
    )

    # Parameters of the Sinkhorn algorithm.
    tau = -0.8  # nesterov-like acceleration
    thresh = 10 ** (-1)  # stopping criterion

    # Elementary operations .....................................................................
    def ave(u, u1):
        "Barycenter subroutine, used by kinetic acceleration through extrapolation."
        return tau * u + (1 - tau) * u1

    def M(u, v):
        "Modified cost for logarithmic updates"
        "$M_{ij} = (-c_{ij} + u_i + v_j) / \epsilon$"
        return (-C + u.unsqueeze(1) + v.unsqueeze(0)) / epsilon

    def lse(A):
        "log-sum-exp"
        return torch.log(
            torch.exp(A).sum(1, keepdim=True) + 1e-6
        )  # add 10^-6 to prevent NaN

    # Actual Sinkhorn loop ......................................................................
    u, v, err = 0.0 * mu, 0.0 * nu, 0.0
    actual_nits = 0  # to check if algorithm terminates because of threshold or max iterations reached

    for _ in range(niter):
        u1 = u  # useful to check the update
        u = epsilon * (torch.log(mu) - lse(M(u, v)).squeeze()) + u
        v = epsilon * (torch.log(nu) - lse(M(u, v).t()).squeeze()) + v
        # accelerated unbalanced iterations
        # u = ave( u, lam * ( epsilon * ( torch.log(mu) - lse(M(u,v)).squeeze()   ) + u ) )
        # v = ave( v, lam * ( epsilon * ( torch.log(nu) - lse(M(u,v).t()).squeeze() ) + v ) )
        err = (u - u1).abs().sum()

        actual_nits += 1
        if (err < thresh).cpu().data.numpy():
            break
    U, V = u, v

    pi = torch.exp(M(U, V))  # Transport plan pi = diag(a)*K*diag(b)
    pi = pi.to("cuda").float()
    return pi  # return the transport


def IPOT(cost_matrix, beta=1, device="cuda"):
    m = cost_matrix.size(0)
    n = cost_matrix.size(1)
    sigma = 1.0 / n * torch.ones([n, 1]).to(device)

    T = torch.ones([m, n]).to(device)
    A = torch.exp(-cost_matrix / beta)

    for t in range(50):
        # BUG: should be elementwise product, * in numpy
        # Q = torch.mm(A, T)
        Q = A * T  # Hardmard product
        for k in range(1):
            delta = 1.0 / (m * torch.mm(Q, sigma))
            sigma = 1.0 / (n * torch.mm(delta.t(), Q)).t()
            # sigma = 1.0 / (n * torch.mv(Q, delta))
        tmp = torch.mm(construct_diag(torch.squeeze(delta)), Q)
        T = torch.mm(tmp, construct_diag(torch.squeeze(sigma)))

    return T


def construct_diag(d):
    n = d.size(0)
    x = torch.zeros([n, n]).to(d.device)
    x[range(n), range(n)] = d.view(-1)
    return x


def get_cost_matrix(x_feature, y_feature):
    C_fea = cost_matrix_cos(x_feature, y_feature)  # Wasserstein cost function
    return C_fea


def cost_matrix_cos(x, y, p=2):
    # return the m*n sized cost matrix
    "Returns the matrix of $|x_i-y_j|^p$."
    # un squeeze differently so that the tensors can broadcast
    # dim-2 (summed over) is the feature dim
    x_col = x.unsqueeze(1)
    y_lin = y.unsqueeze(0)

    cos = nn.CosineSimilarity(dim=2, eps=1e-6)
    c = torch.clamp(1 - cos(x_col, y_lin), min=0)

    return c



# (c) 2024 LAiSR-SK
# This code is licensed under the MIT license (see LICENSE.md).

from airtk.defense.feature_scatter.fs_train import (
    FeatureScatterTraining,
)

__all__ = ["FeatureScatterTraining"]



import torch

from ..attack import Attack


class MultiAttack(Attack):
    r"""
    MultiAttack is a class to attack a model with various attacks agains same images and labels.

    Arguments:
        model (nn.Module): model to attack.
        attacks (list): list of attacks.

    Examples::
        >>> atk1 = torchattacks.PGD(model, eps=8/255, alpha=2/255, iters=40, random_start=True)
        >>> atk2 = torchattacks.PGD(model, eps=8/255, alpha=2/255, iters=40, random_start=True)
        >>> atk = torchattacks.MultiAttack([atk1, atk2])
        >>> adv_images = attack(images, labels)

    """

    def __init__(self, attacks, verbose=False):
        super().__init__("MultiAttack", attacks[0].model)
        self.attacks = attacks
        self.verbose = verbose
        self.supported_mode = ["default"]

        self.check_validity()

        self._accumulate_multi_atk_records = False
        self._multi_atk_records = [0.0]

    def check_validity(self):
        if len(self.attacks) < 2:
            raise ValueError("More than two attacks should be given.")

        ids = [id(attack.model) for attack in self.attacks]
        if len(set(ids)) != 1:
            raise ValueError(
                "At least one of attacks is referencing a different model."
            )

    def forward(self, images, labels):
        r"""
        Overridden.
        """
        batch_size = images.shape[0]
        fails = torch.arange(batch_size).to(self.device)
        final_images = images.clone().detach().to(self.device)
        labels = labels.clone().detach().to(self.device)

        multi_atk_records = [batch_size]

        for _, attack in enumerate(self.attacks):
            adv_images = attack(images[fails], labels[fails])

            outputs = self.get_logits(adv_images)
            _, pre = torch.max(outputs.data, 1)

            corrects = pre == labels[fails]
            wrongs = ~corrects

            succeeds = torch.masked_select(fails, wrongs)
            succeeds_of_fails = torch.masked_select(
                torch.arange(fails.shape[0]).to(self.device), wrongs
            )

            final_images[succeeds] = adv_images[succeeds_of_fails]

            fails = torch.masked_select(fails, corrects)
            multi_atk_records.append(len(fails))

            if len(fails) == 0:
                break

        if self.verbose:
            print(self._return_sr_record(multi_atk_records))

        if self._accumulate_multi_atk_records:
            self._update_multi_atk_records(multi_atk_records)

        return final_images

    def _clear_multi_atk_records(self):
        self._multi_atk_records = [0.0]

    def _covert_to_success_rates(self, multi_atk_records):
        sr = [
            ((1 - multi_atk_records[i] / multi_atk_records[0]) * 100)
            for i in range(1, len(multi_atk_records))
        ]
        return sr

    def _return_sr_record(self, multi_atk_records):
        sr = self._covert_to_success_rates(multi_atk_records)
        return "Attack success rate: " + " | ".join(
            ["%2.2f %%" % item for item in sr]
        )

    def _update_multi_atk_records(self, multi_atk_records):
        for i, item in enumerate(multi_atk_records):
            self._multi_atk_records[i] += item

    def save(
        self,
        data_loader,
        save_path=None,
        verbose=True,
        return_verbose=False,
        save_predictions=False,
        save_clean_images=False,
    ):
        r"""
        Overridden.
        """
        self._clear_multi_atk_records()
        prev_verbose = self.verbose
        self.verbose = False
        self._accumulate_multi_atk_records = True

        for i, attack in enumerate(self.attacks):
            self._multi_atk_records.append(0.0)

        if return_verbose:
            rob_acc, l2, elapsed_time = super().save(
                data_loader,
                save_path,
                verbose,
                return_verbose,
                save_predictions,
                save_clean_images,
            )
            sr = self._covert_to_success_rates(self._multi_atk_records)
        elif verbose:
            super().save(
                data_loader,
                save_path,
                verbose,
                return_verbose,
                save_predictions,
                save_clean_images,
            )
            sr = self._covert_to_success_rates(self._multi_atk_records)
        else:
            super().save(
                data_loader,
                save_path,
                False,
                False,
                save_predictions,
                save_clean_images,
            )

        self._clear_multi_atk_records()
        self._accumulate_multi_atk_records = False
        self.verbose = prev_verbose

        if return_verbose:
            return rob_acc, sr, l2, elapsed_time

    def _save_print(self, progress, rob_acc, l2, elapsed_time, end):
        r"""
        Overridden.
        """
        print(
            "- Save progress: %2.2f %% / Robust accuracy: %2.2f %%"
            % (progress, rob_acc)
            + " / "
            + self._return_sr_record(self._multi_atk_records)
            + " / L2: %1.5f (%2.3f it/s) \t" % (l2, elapsed_time),
            end=end,
        )



import os
import sys
from abc import ABCMeta, abstractmethod, abstractproperty
from typing import List

import torch


class TrainingConfigBase(metaclass=ABCMeta):
    """
    Base class for training
    """

    # directory handling
    @property
    def abs_current_dir(self):
        return os.path.realpath("./")

    @property
    def log_dir(self):
        if not os.path.exists("./log"):
            os.mkdir("./log")
        return os.path.join(self.abs_current_dir, "log")

    @property
    def model_dir(self):
        log_dir = self.log_dir
        model_dir = os.path.join(log_dir, "models")
        # print(model_dir)
        if not os.path.exists(model_dir):
            os.mkdir(model_dir)
        return model_dir

    @abstractproperty
    def lib_dir(self):
        pass

    # training setting
    @abstractproperty
    def num_epochs(self):
        pass

    @property
    def val_interval(self):
        """
        Specify how many epochs between two validation steps
        Return <= 0 means no validation phase
        """
        return 0

    @abstractmethod
    def create_optimizer(self, params) -> torch.optim.Optimizer:
        """
        params (iterable): iterable of parameters to optimize or dicts defining
                           parameter groups
        """
        pass

    @abstractmethod
    def create_lr_scheduler(
        self, optimizer: torch.optim.Optimizer
    ) -> torch.optim.lr_scheduler._LRScheduler:
        pass

    @abstractmethod
    def create_loss_function(self) -> torch.nn.modules.loss._Loss:
        pass

    def create_attack_method(self, *inputs):
        """
        Perform adversarial training against xxx adversary
        Return None means natural training
        """
        return None

    # Evaluation Setting

    def create_evaluation_attack_method(self, *inputs):
        """
        evaluating the robustness of model against xxx adversary
        Return None means only measuring clean accuracy
        """
        return None


class SGDOptimizerMaker:
    def __init__(self, lr=0.1, momentum=0.9, weight_decay=1e-4):
        self.lr = lr
        self.momentum = momentum
        self.weight_decay = weight_decay

    def __call__(self, params):
        return torch.optim.SGD(
            params,
            lr=self.lr,
            momentum=self.momentum,
            weight_decay=self.weight_decay,
        )


class PieceWiseConstantLrSchedulerMaker:
    def __init__(self, milestones: List[int], gamma: float = 0.1):
        self.milestones = milestones
        self.gamma = gamma

    def __call__(self, optimizer):
        return torch.optim.lr_scheduler.MultiStepLR(
            optimizer, milestones=self.milestones, gamma=self.gamma
        )


class IPGDAttackMethodMaker:
    def __init__(self, eps, sigma, nb_iters, norm, mean, std):
        self.eps = eps
        self.sigma = sigma
        self.nb_iters = nb_iters
        self.norm = norm
        self.mean = mean
        self.std = std

    def __call__(self, DEVICE):
        father_dir = os.path.join(
            "/", *os.path.realpath(__file__).split(os.path.sep)[:-2]
        )
        # print(father_dir)
        if father_dir not in sys.path:
            sys.path.append(father_dir)
        from airtk.defense.yopo.attack_pgd import IPGD

        return IPGD(
            self.eps,
            self.sigma,
            self.nb_iters,
            self.norm,
            DEVICE,
            self.mean,
            self.std,
        )


class LambdaLrSchedulerMaker:
    def __init__(self, func, last_epoch=-1):
        assert callable(func)

        self.func = func
        self.last_epoch = last_epoch

    def __call__(self, parameters):
        from torch.optim.lr_scheduler import LambdaLR

        lr_schduler = LambdaLR(parameters, self.func, self.last_epoch)
        return lr_schduler



# Modified from https://github.com/Harry24k/pytorch-custom-utils
from collections.abc import Iterable
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.pylab as pylab
import torch


def init_settings(
    font_size=None,
    title_size=None,
    label_size=None,
    xlabel_size=None,
    ylabel_size=None,
    legend_size=None,
):
    params = {
        "font.size": font_size,
        "axes.titlesize": title_size,
        "axes.labelsize": label_size,
        "xtick.labelsize": xlabel_size,
        "ytick.labelsize": ylabel_size,
        "legend.fontsize": legend_size,
    }
    _del_none(params)
    pylab.rcParams.update(params)
    print("rcParams updated.")


def init_plot(
    ax=None,
    figsize=(6, 6),
    title=None,
    xlabel=None,
    ylabel=None,
    xlim=None,
    ylim=None,
    pad_ratio=0,
    show_axis=True,
    show_grid=False,
    tight=True,
):
    if ax is None:
        ax = plt.subplots(1, 1, figsize=figsize)[1]

    if title is not None:
        ax.set_title(title)

    if xlabel is not None:
        ax.set_xlabel(xlabel)

    if ylabel is not None:
        ax.set_ylabel(ylabel)

    if xlim is not None:
        xlim = _add_margin(*xlim, pad_ratio)
        ax.set_xlim(xlim)

    if ylim is not None:
        ylim = _add_margin(*ylim, pad_ratio)
        ax.set_ylim(ylim)

    if not show_axis:
        ax.axis("off")

    if show_grid:
        ax.grid()

    if tight:
        plt.tight_layout()

    return ax


def make_twin(ax, ylabel=None, ylim=None, pad_ratio=0):
    ax2 = ax.twinx()

    if ylabel is not None:
        ax2.set_ylabel(ylabel)

    if ylim is not None:
        ylim = _add_margin(*ylim, pad_ratio)
        ax2.set_ylim(ylim)

    return ax2


def plot_line(
    ax,
    input,
    x=None,
    linewidth=None,
    linestyle=None,
    color=None,
    label=None,
    alpha=None,
    dashes=None,
    marker=None,
    markerfacecolor=None,
    markersize=None,
):
    kwargs = {
        "linewidth": linewidth,
        "linestyle": linestyle,
        "color": color,
        "label": label,
        "alpha": alpha,
        "dashes": dashes,
        "marker": marker,
        "markerfacecolor": markerfacecolor,
        "markersize": markersize,
    }
    input = _to_numpy(input)
    _del_none(kwargs)

    if x is None:
        x = list(range(len(input) + 1))[1:]

    ax.plot(x, input, **kwargs)


def _add_margin(x_min, x_max, ratio=0.3):
    range = x_max - x_min
    mean = (x_max + x_min) / 2
    return mean - range / 2 * (1 + ratio), mean + range / 2 * (1 + ratio)


def _del_none(input):
    keys = input.copy().keys()
    for key in keys:
        if input[key] is None:
            del input[key]


def _to_numpy(input):
    if isinstance(input, np.ndarray):
        return input
    elif isinstance(input, pd.core.series.Series):
        return input.values
    elif isinstance(input, pd.core.frame.DataFrame):
        return input.values
    elif isinstance(input, torch.Tensor):
        return input.detach().cpu().numpy()
    else:
        return input


#         raise RuntimeError("Please input tensor or numpy array.")


def _to_array(inputs, lengths):
    for i, input in enumerate(inputs):
        if not isinstance(input, Iterable):
            inputs[i] = [input] * lengths
    return inputs



import json
import math
import os
import sys
from typing import List

import torch


def torch_accuracy(output, target, topk=(1,)) -> List[torch.Tensor]:
    """
    param output, target: should be torch Variable
    """
    # assert isinstance(output, torch.cuda.Tensor), 'expecting Torch Tensor'
    # assert isinstance(target, torch.Tensor), 'expecting Torch Tensor'
    # print(type(output))

    topn = max(topk)
    batch_size = output.size(0)

    _, pred = output.topk(topn, 1, True, True)
    pred = pred.t()

    is_correct = pred.eq(target.view(1, -1).expand_as(pred))

    ans = []
    for i in topk:
        is_correct_i = is_correct[:i].view(-1).float().sum(0, keepdim=True)
        ans.append(is_correct_i.mul_(100.0 / batch_size))

    return ans


class AvgMeter:
    """
    Computing mean
    """

    name = "No name"

    def __init__(self, name="No name"):
        self.name = name
        self.reset()

    def reset(self):
        self.sum = 0
        self.mean = 0
        self.num = 0
        self.now = 0

    def update(self, mean_var, count=1):
        if math.isnan(mean_var):
            mean_var = 1e6
            print("Avgmeter getting Nan!")
        self.now = mean_var
        self.num += count

        self.sum += mean_var * count
        self.mean = float(self.sum) / self.num


def save_args(args, save_dir=None):
    if save_dir == None:
        param_path = os.path.join(args.resume, "params.json")
    else:
        param_path = os.path.join(save_dir, "params.json")

    # logger.info("[*] MODEL dir: %s" % args.resume)
    # logger.info("[*] PARAM path: %s" % param_path)

    with open(param_path, "w") as fp:
        json.dump(args.__dict__, fp, indent=4, sort_keys=True)


def mkdir(path):
    if not os.path.exists(path):
        print(f"creating dir {path}")
        os.mkdir(path)


def save_checkpoint(now_epoch, net, optimizer, lr_scheduler, file_name):
    checkpoint = {
        "epoch": now_epoch,
        "state_dict": net.state_dict(),
        "optimizer_state_dict": optimizer.state_dict(),
        "lr_scheduler_state_dict": lr_scheduler.state_dict(),
    }
    if os.path.exists(file_name):
        print(f"Overwriting {file_name}")
    torch.save(checkpoint, file_name)
    link_name = os.path.join(
        "/", *file_name.split(os.path.sep)[:-1], "last.checkpoint"
    )
    # print(link_name)
    # make_symlink(source = file_name, link_name=link_name)


def load_checkpoint(file_name, net=None, optimizer=None, lr_scheduler=None):
    if os.path.isfile(file_name):
        print(f"=> loading checkpoint '{file_name}'")
        check_point = torch.load(file_name)
        if net is not None:
            print("Loading network state dict")
            net.load_state_dict(check_point["state_dict"])
        if optimizer is not None:
            print("Loading optimizer state dict")
            optimizer.load_state_dict(check_point["optimizer_state_dict"])
        if lr_scheduler is not None:
            print("Loading lr_scheduler state dict")
            lr_scheduler.load_state_dict(
                check_point["lr_scheduler_state_dict"]
            )

        return check_point["epoch"]
    else:
        print(f"=> no checkpoint found at '{file_name}'")


def make_symlink(source, link_name):
    """
    Note: overwriting enabled!
    """
    if os.path.exists(link_name):
        # print("Link name already exist! Removing '{}' and overwriting".format(link_name))
        os.remove(link_name)
    if os.path.exists(source):
        os.symlink(source, link_name)
        return
    else:
        print("Source path not exists")
    # print('SymLink Wrong!')


def add_path(path):
    if path not in sys.path:
        print(f"Adding {path}")
        sys.path.append(path)



# (c) 2024 LAiSR-SK
# This code is licensed under the MIT license (see LICENSE.md).

import os
import time
from collections import namedtuple
from os import PathLike

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from torch.autograd import Variable
from torchvision import datasets, transforms

from airtk.defense.tradesawp.trades_awp_utils_awp import (
    TradesAWP,
)
from airtk.defense.tradesawp.trades_awp_utils_eval import (
    accuracy,
)
from airtk.defense.tradesawp.trades_awp_utils_logger import (
    Logger,
)
from airtk.defense.tradesawp.trades_awp_utils_misc import (
    AverageMeter,
)

# from utils import Bar, Logger, AverageMeter, accuracy
# import models
from airtk.model import WideResNet


class TradesawpTraining:
    def __init__(
        self,
        dataset: str,
        model: str,
        batch_size: int = 128,
        epochs: int = 200,
        weight_decay: float = 5e-4,
        lr: float = 0.1,
        momentum: float = 0.9,
        norm: str = "l_inf",
        epsilon: int = 8,
        num_steps: int = 10,
        step_size: int = 2,
        beta: float = 6.0,
        seed: int = 1,
        save_freq: int = 1,
        awp_gamma: float = 0.005,
        awp_warmup: int = 10,
        model_dir: PathLike = "model/data",
    ) -> None:
        ArgsPrototype = namedtuple(
            "ArgsPrototype",
            [
                "arch",
                "batch_size",
                "test_batch_size",
                "epochs",
                "start_epoch",
                "data",
                "data_path",
                "weight_decay",
                "lr",
                "momentum",
                "no_cuda",
                "norm",
                "epsilon",
                "num_steps",
                "step_size",
                "beta",
                "seed",
                "model_dir",
                "resume_model",
                "resume_optim",
                "save_freq",
                "awp_gamma",
                "awp_warmup",
            ],
        )
        self._args = ArgsPrototype(
            model,
            batch_size,
            batch_size,
            epochs,
            1,
            dataset,
            "data",
            weight_decay,
            lr,
            momentum,
            False,
            norm,
            epsilon,
            num_steps,
            step_size,
            beta,
            seed,
            model_dir,
            model_dir,
            model_dir,
            save_freq,
            awp_gamma,
            awp_warmup,
        )

    def __call__(self) -> None:
        torch.manual_seed(self._args.seed)
        main_tradesawp(self._args)


# parser = argparse.ArgumentParser(
# description="PyTorch CIFAR TRADES Adversarial Training"
# )
# parser.add_argument("--arch", type=str, default="WideResNet34")
# parser.add_argument(
# "--batch-size",
# type=int,
# default=128,
# metavar="N",
# help="input batch size for training (default: 128)",
# )
# parser.add_argument(
# "--test-batch-size",
# type=int,
# default=128,
# metavar="N",
# help="input batch size for testing (default: 128)",
# )
# parser.add_argument(
# "--epochs",
# type=int,
# default=200,
# metavar="N",
# help="number of epochs to train",
# )
# parser.add_argument(
# "--start_epoch",
# type=int,
# default=1,
# metavar="N",
# help="retrain from which epoch",
# )
# parser.add_argument(
# "--data", type=str, default="CIFAR10", choices=["CIFAR10", "CIFAR100"]
# )
# parser.add_argument(
# "--data-path",
# type=str,
# default="../data",
# help="where is the dataset CIFAR-10",
# )
# parser.add_argument(
# "--weight-decay", "--wd", default=5e-4, type=float, metavar="W"
# )
# parser.add_argument(
# "--lr", type=float, default=0.1, metavar="LR", help="learning rate"
# )
# parser.add_argument(
# "--momentum", type=float, default=0.9, metavar="M", help="SGD momentum"
# )
# parser.add_argument(
# "--no-cuda",
# action="store_true",
# default=False,
# help="disables CUDA training",
# )
# parser.add_argument(
# "--norm",
# default="l_inf",
# type=str,
# choices=["l_inf", "l_2"],
# help="The threat model",
# )
# parser.add_argument("--epsilon", default=8, type=float, help="perturbation")
# parser.add_argument(
# "--num-steps", default=10, type=int, help="perturb number of steps"
# )
# parser.add_argument(
# "--step-size", default=2, type=float, help="perturb step size"
# )
# parser.add_argument(
# "--beta",
# default=6.0,
# type=float,
# help="regularization, i.e., 1/lambda in TRADES",
# )
# parser.add_argument(
# "--seed", type=int, default=1, metavar="S", help="random seed (default: 1)"
# )
# parser.add_argument(
# "--model-dir",
# default="./data/model",
# help="directory of model for saving checkpoint",
# )
# parser.add_argument(
# "--resume-model",
# default="",
# type=str,
# help="directory of model for retraining",
# )
# parser.add_argument(
# "--resume-optim",
# default="",
# type=str,
# help="directory of optimizer for retraining",
# )
# parser.add_argument(
# "--save-freq",
# "-s",
# default=1,
# type=int,
# metavar="N",
# help="save frequency",
# )

# parser.add_argument(
# "--awp-gamma",
# default=0.005,
# type=float,
# help="whether or not to add parametric noise",
# )
# parser.add_argument(
# "--awp-warmup",
# default=10,
# type=int,
# help="We could apply AWP after some epochs for accelerating.",
# )
# args = parser.parse_args()


def setup(data, args):
    args.data = data
    epsilon = args.epsilon / 255
    step_size = args.step_size / 255
    if args.awp_gamma <= 0.0:
        args.awp_warmup = np.infty
    NUM_CLASSES = 100 if args.data == "CIFAR100" else 10
    return args, epsilon, step_size, NUM_CLASSES


# settings
# model_dir = args.model_dir
# if not os.path.exists(model_dir):
#    os.makedirs(model_dir)
# torch.manual_seed(args.seed)
# device = torch.device("cuda" if use_cuda else "cpu")
# kwargs = {"num_workers": 2, "pin_memory": True} if use_cuda else {}

# setup data loader
# transform_train = transforms.Compose(
# [
# transforms.RandomCrop(32, padding=4),
# transforms.RandomHorizontalFlip(),
# transforms.ToTensor(),
# ]
# )
# transform_test = transforms.Compose(
# [
# transforms.ToTensor(),
# ]
# )
# trainset = getattr(datasets, args.data)(
# root=args.data_path, train=True, download=True, transform=transform_train
# )
# testset = getattr(datasets, args.data)(
# root=args.data_path, train=False, download=True, transform=transform_test
# )
# train_loader = torch.utils.data.DataLoader(
# trainset, batch_size=args.batch_size, shuffle=True, **kwargs
# )
# test_loader = torch.utils.data.DataLoader(
# testset, batch_size=args.test_batch_size, shuffle=False, **kwargs
# )


def perturb_input(
    model,
    x_natural,
    step_size=0.003,
    epsilon=0.031,
    perturb_steps=10,
    distance="l_inf",
):
    device = "cuda" if torch.cuda.is_available() else "cpu"
    model.eval()
    batch_size = len(x_natural)
    if distance == "l_inf":
        x_adv = (
            x_natural.detach()
            + 0.001 * torch.randn(x_natural.shape).to(device).detach()
        )
        for _ in range(perturb_steps):
            x_adv.requires_grad_()
            with torch.enable_grad():
                loss_kl = F.kl_div(
                    F.log_softmax(model(x_adv), dim=1),
                    F.softmax(model(x_natural), dim=1),
                    reduction="sum",
                )
            grad = torch.autograd.grad(loss_kl, [x_adv])[0]
            x_adv = x_adv.detach() + step_size * torch.sign(grad.detach())
            x_adv = torch.min(
                torch.max(x_adv, x_natural - epsilon), x_natural + epsilon
            )
            x_adv = torch.clamp(x_adv, 0.0, 1.0)
    elif distance == "l_2":
        delta = 0.001 * torch.randn(x_natural.shape).to(device).detach()
        delta = Variable(delta.data, requires_grad=True)

        # Setup optimizers
        optimizer_delta = optim.SGD([delta], lr=epsilon / perturb_steps * 2)

        for _ in range(perturb_steps):
            adv = x_natural + delta

            # optimize
            optimizer_delta.zero_grad()
            with torch.enable_grad():
                loss = (-1) * F.kl_div(
                    F.log_softmax(model(adv), dim=1),
                    F.softmax(model(x_natural), dim=1),
                    reduction="sum",
                )
            loss.backward()
            # renorming gradient
            grad_norms = delta.grad.view(batch_size, -1).norm(p=2, dim=1)
            delta.grad.div_(grad_norms.view(-1, 1, 1, 1))
            # avoid nan or inf if gradient is 0
            # if (grad_norms == 0).any():
            #     delta.grad[grad_norms == 0] = torch.randn_like(delta.grad[grad_norms == 0])
            optimizer_delta.step()

            # projection
            delta.data.add_(x_natural)
            delta.data.clamp_(0, 1).sub_(x_natural)
            delta.data.renorm_(p=2, dim=0, maxnorm=epsilon)
        x_adv = Variable(x_natural + delta, requires_grad=False)
    else:
        x_adv = (
            x_natural.detach()
            + 0.001 * torch.randn(x_natural.shape).to(device).detach()
        )
        x_adv = torch.clamp(x_adv, 0.0, 1.0)
    return x_adv


def train(model, train_loader, optimizer, epoch, awp_adversary, data):
    args, epsilon, step_size, NUM_CLASSES = setup(data)
    batch_time = AverageMeter()
    losses = AverageMeter()
    top1 = AverageMeter()
    end = time.time()

    print(f"epoch: {epoch}")
    # bar = Bar('Processing', max=len(train_loader))

    device = "cuda" if torch.cuda.is_available() else "cpu"
    for batch_idx, (data, target) in enumerate(train_loader):
        x_natural, target = data.to(device), target.to(device)

        # craft adversarial examples
        x_adv = perturb_input(
            model=model,
            x_natural=x_natural,
            step_size=step_size,
            epsilon=epsilon,
            perturb_steps=args.num_steps,
            distance=args.norm,
        )

        model.train()
        # calculate adversarial weight perturbation
        if epoch >= args.awp_warmup:
            awp = awp_adversary.calc_awp(
                inputs_adv=x_adv,
                inputs_clean=x_natural,
                targets=target,
                beta=args.beta,
            )
            awp_adversary.perturb(awp)

        optimizer.zero_grad()
        logits_adv = model(x_adv)
        loss_robust = F.kl_div(
            F.log_softmax(logits_adv, dim=1),
            F.softmax(model(x_natural), dim=1),
            reduction="batchmean",
        )
        # calculate natural loss and backprop
        logits = model(x_natural)
        loss_natural = F.cross_entropy(logits, target)
        loss = loss_natural + args.beta * loss_robust

        prec1, prec5 = accuracy(logits_adv, target, topk=(1, 5))
        losses.update(loss.item(), x_natural.size(0))
        top1.update(prec1.item(), x_natural.size(0))

        # update the parameters at last
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

        if epoch >= args.awp_warmup:
            awp_adversary.restore(awp)

        batch_time.update(time.time() - end)
        end = time.time()

        """bar.suffix = '({batch}/{size}) Batch: {bt:.3f}s| Total:{total:}| ETA:{eta:}| Loss:{loss:.4f}| top1:{top1:.2f}'.format(
            batch=batch_idx + 1,
            size=len(train_loader),
            bt=batch_time.val,
            total=bar.elapsed_td,
            eta=bar.eta_td,
            loss=losses.avg,
            top1=top1.avg)
        bar.next()"""
    # bar.finish()
    return losses.avg, top1.avg


def test(model, test_loader, criterion):
    global best_acc
    model.eval()

    batch_time = AverageMeter()
    losses = AverageMeter()
    top1 = AverageMeter()
    end = time.time()

    # bar = Bar('Processing', max=len(test_loader))
    device = "cuda" if torch.cuda.is_available() else "cpu"
    with torch.no_grad():
        for batch_idx, (inputs, targets) in enumerate(test_loader):
            inputs, targets = inputs.to(device), targets.to(device)
            outputs = model(inputs)
            loss = criterion(outputs, targets)

            prec1, prec5 = accuracy(outputs.data, targets.data, topk=(1, 5))
            losses.update(loss.item(), inputs.size(0))
            top1.update(prec1.item(), inputs.size(0))

            batch_time.update(time.time() - end)
            end = time.time()

            """bar.suffix = '({batch}/{size}) Batch: {bt:.3f}s| Total: {total:}| ETA: {eta:}| Loss:{loss:.4f}| top1: {top1:.2f}'.format(
                batch=batch_idx + 1,
                size=len(test_loader),
                bt=batch_time.avg,
                total=bar.elapsed_td,
                eta=bar.eta_td,
                loss=losses.avg,
                top1=top1.avg)
            bar.next()"""
    # bar.finish()
    return losses.avg, top1.avg


def adjust_learning_rate(optimizer, epoch, args):
    """decrease the learning rate"""
    lr = args.lr
    if epoch >= 100:
        lr = args.lr * 0.1
    if epoch >= 150:
        lr = args.lr * 0.01
    for param_group in optimizer.param_groups:
        param_group["lr"] = lr
    return lr


def main(args):
    kwargs = (
        {"num_workers": 2, "pin_memory": True}
        if torch.cuda.is_available()
        else {}
    )

    # setup data loader
    transform_train = transforms.Compose(
        [
            transforms.RandomCrop(32, padding=4),
            transforms.RandomHorizontalFlip(),
            transforms.ToTensor(),
        ]
    )
    transform_test = transforms.Compose(
        [
            transforms.ToTensor(),
        ]
    )
    trainset = getattr(datasets, args.data)(
        root=args.data_path,
        train=True,
        download=True,
        transform=transform_train,
    )
    testset = getattr(datasets, args.data)(
        root=args.data_path,
        train=False,
        download=True,
        transform=transform_test,
    )
    train_loader = torch.utils.data.DataLoader(
        trainset, batch_size=args.batch_size, shuffle=True, **kwargs
    )
    test_loader = torch.utils.data.DataLoader(
        testset, batch_size=args.test_batch_size, shuffle=False, **kwargs
    )
    data = args.data
    args, epsilon, step_size, NUM_CLASSES = setup(data)
    # init model, ResNet18() can be also used here for training
    device = "cuda" if torch.cuda.is_available() else "cpu"
    # model= ResNet18(num_classes=NUM_CLASSES).to(device)
    model = WideResNet(
        depth=34, num_classes=100 if data == "cifar100" else 10
    ).to(device)
    # model = nn.DataParallel(getattr(models, args.arch)(num_classes=NUM_CLASSES)).to(device)
    optimizer = optim.SGD(
        model.parameters(),
        lr=args.lr,
        momentum=args.momentum,
        weight_decay=args.weight_decay,
    )

    # We use a proxy model to calculate AWP, which does not affect the statistics of BN.
    # proxy = nn.DataParallel(getattr(models, args.arch)(num_classes=NUM_CLASSES)).to(device)
    # proxy = ResNet18(num_classes=NUM_CLASSES)
    proxy = WideResNet(
        depth=34, num_classes=100 if data == "cifar100" else 10
    ).to(device)
    proxy_optim = optim.SGD(proxy.parameters(), lr=args.lr)
    awp_adversary = TradesAWP(
        model=model, proxy=proxy, proxy_optim=proxy_optim, gamma=args.awp_gamma
    )

    criterion = nn.CrossEntropyLoss()

    logger = Logger(os.path.join(args.model_dir, "log.txt"), title=args.arch)
    logger.set_names(
        [
            "Learning Rate",
            "Adv Train Loss",
            "Nat Train Loss",
            "Nat Val Loss",
            "Adv Train Acc.",
            "Nat Train Acc.",
            "Nat Val Acc.",
        ]
    )

    if args.resume_model:
        model.load_state_dict(
            torch.load(args.resume_model, map_location=device)
        )
    if args.resume_optim:
        optimizer.load_state_dict(
            torch.load(args.resume_optim, map_location=device)
        )

    for epoch in range(args.start_epoch, args.epochs + 1):
        # adjust learning rate for SGD
        lr = adjust_learning_rate(optimizer, epoch, args)

        # adversarial training
        adv_loss, adv_acc = train(
            model, train_loader, optimizer, epoch, awp_adversary, data
        )

        # evaluation on natural examples
        print(
            "================================================================"
        )
        train_loss, train_acc = test(model, train_loader, criterion)
        val_loss, val_acc = test(model, test_loader, criterion)
        print(
            "================================================================"
        )

        logger.append(
            [lr, adv_loss, train_loss, val_loss, adv_acc, train_acc, val_acc]
        )

        # save checkpoint
        if epoch % args.save_freq == 0:
            torch.save(
                model.state_dict(),
                os.path.join(args.model_dir, f"ours-model-epoch{epoch}.pt"),
            )
            torch.save(
                optimizer.state_dict(),
                os.path.join(
                    args.model_dir, f"ours-opt-checkpoint_epoch{epoch}.tar"
                ),
            )
    return model


def main_trades_awp_10(args):
    model = main(args)

    md = "./data/model"
    torch.save(
        model.state_dict(),
        os.path.join(
            md, "model-trades-awp-{}-{}.pt".format("cifar10", "wideres34")
        ),
    )

    # target_model = ResNet18(num_classes=10)
    device = "cuda" if torch.cuda.is_available() else "cpu"
    target_model = WideResNet(depth=34, num_classes=10).to(device)
    from collections import OrderedDict

    checkpoint = torch.load(
        os.path.join(
            md, "model-trades-awp-{}-{}.pt".format("cifar10", "wideres34")
        )
    )
    try:
        target_model.load_state_dict(checkpoint)
    except:
        new_state_dict = OrderedDict()
        for k, v in checkpoint.items():
            name = k[7:]  # remove `module.`
            new_state_dict[name] = v
        target_model.load_state_dict(new_state_dict, False)

    target_model = torch.nn.DataParallel(target_model).cuda()

    torch.save(
        target_model.module.state_dict(),
        os.path.join(
            args.model_dir,
            "model-trades-awp-{}-{}.pt".format("cifar10", "wideres34"),
        ),
    )


def main_trades_awp_100(args):
    model = main(args)

    md = "./data/model"
    torch.save(
        model.state_dict(),
        os.path.join(
            md, "model-trades-awp-{}-{}.pt".format("cifar100", "wideres34")
        ),
    )

    # target_model = ResNet18(num_classes=100)
    device = "cuda" if torch.cuda.is_available() else "cpu"
    target_model = WideResNet(depth=34, num_classes=100).to(device)

    from collections import OrderedDict

    checkpoint = torch.load(
        os.path.join(
            md, "model-trades-awp-{}-{}.pt".format("cifar100", "wideres34")
        )
    )
    try:
        target_model.load_state_dict(checkpoint)
    except:
        new_state_dict = OrderedDict()
        for k, v in checkpoint.items():
            name = k[7:]  # remove `module.`
            new_state_dict[name] = v
        target_model.load_state_dict(new_state_dict, False)

    target_model = torch.nn.DataParallel(target_model).cuda()

    torch.save(
        target_model.module.state_dict(),
        os.path.join(
            args.model_dir,
            "model-trades-awp-{}-{}.pt".format("cifar100", "wideres34"),
        ),
    )


def main_tradesawp(args):
    if args.data == "cifar10":
        main_trades_awp_10(args)
    elif args.data == "cifar100":
        main_trades_awp_100(args)
    else:
        raise NotImplementedError



import torch
import torch.nn as nn
import torch.nn.functional as F

from ...attacks import PGD

from .advtrainer import AdvTrainer


class MART(AdvTrainer):
    r"""
    MART in 'Improving Adversarial Robustness Requires Revisiting Misclassified Examples'
    [https://openreview.net/forum?id=rklOg6EFwS]
    [https://github.com/YisenWang/MART]

    Attributes:
        self.rmodel : rmodel.
        self.device : device where rmodel is.
        self.optimizer : optimizer.
        self.scheduler : scheduler (Automatically updated).
        self.curr_epoch : current epoch starts from 1 (Automatically updated).
        self.curr_iter : current iters starts from 1 (Automatically updated).

    Arguments:
        rmodel (nn.Module): rmodel to train.
        eps (float): strength of the attack or maximum perturbation.
        alpha (float): step size.
        steps (int): number of steps.
        beta (float): trade-off regularization parameter.
        random_start (bool): using random initialization of delta.
    """

    def __init__(self, rmodel, eps, alpha, steps, beta, random_start=True):
        super().__init__(rmodel)
        self.atk = PGD(rmodel, eps, alpha, steps, random_start)
        self.beta = beta

    def calculate_cost(self, train_data, reduction="mean"):
        r"""
        Overridden.
        """
        images, labels = train_data
        images = images.to(self.device)
        labels = labels.to(self.device)

        adv_images = self.atk(images, labels)

        logits_clean = self.rmodel(images)
        logits_adv = self.rmodel(adv_images)

        probs_adv = F.softmax(logits_adv, dim=1)

        # Caculate BCELoss
        tmp1 = torch.argsort(probs_adv, dim=1)[:, -2:]
        new_y = torch.where(tmp1[:, -1] == labels, tmp1[:, -2], tmp1[:, -1])
        loss_bce_adv = F.cross_entropy(
            logits_adv, labels, reduction="none"
        ) + F.nll_loss(torch.log(1.0001 - probs_adv + 1e-12), new_y, reduction="none")

        # Caculate KLLoss
        probs_clean = F.softmax(logits_clean, dim=1)
        log_prob_adv = torch.log(probs_adv + 1e-12)
        loss_kl = torch.sum(
            nn.KLDivLoss(reduction="none")(log_prob_adv, probs_clean), dim=1
        )
        true_probs = torch.gather(
            probs_clean, 1, (labels.unsqueeze(1)).long()
        ).squeeze()
        loss_weighted_kl = loss_kl * (1.0000001 - true_probs)

        cost = loss_bce_adv + self.beta * loss_weighted_kl

        self.add_record_item("Loss", cost.mean().item())
        self.add_record_item("BALoss", loss_bce_adv.mean().item())
        self.add_record_item("WKLoss", loss_weighted_kl.mean().item())

        return cost.mean() if reduction == "mean" else cost



# (c) 2024 LAiSR-SK

# This code is licensed under the MIT license (see LICENSE.md).
from airtk.defense.atawp.at_awp_train_cifar10 import (
    main_atawp10,
)
from airtk.defense.atawp.at_awp_train_cifar100 import (
    main_atawp100,
)


def main_atawp(ds_name):
    if ds_name == "cifar10":
        main_atawp10()
    elif ds_name == "cifar100":
        main_atawp100()
    else:
        raise NotImplementedError


if __name__ == "__main__":
    main_atawp("cifar10")



# (c) 2024 LAiSR-SK
# This code is licensed under the MIT license (see LICENSE.md).
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from torch.autograd import Variable


def squared_l2_norm(x):
    flattened = x.view(x.unsqueeze(0).shape[0], -1)
    return (flattened**2).sum(1)


def l2_norm(x):
    return squared_l2_norm(x).sqrt()


def trades_loss(
    model,
    x_natural,
    y,
    optimizer,
    step_size=0.003,
    epsilon=0.031,
    perturb_steps=10,
    beta=1.0,
    distance="l_inf",
):
    # define KL-loss
    criterion_kl = nn.KLDivLoss(size_average=False)
    model.eval()
    batch_size = len(x_natural)
    # generate adversarial example
    x_adv = (
        x_natural.detach()
        + 0.001 * torch.randn(x_natural.shape).cuda().detach()
    )
    if distance == "l_inf":
        for _ in range(perturb_steps):
            x_adv.requires_grad_()
            with torch.enable_grad():
                loss_kl = criterion_kl(
                    F.log_softmax(model(x_adv), dim=1),
                    F.softmax(model(x_natural), dim=1),
                )
            grad = torch.autograd.grad(loss_kl, [x_adv])[0]
            x_adv = x_adv.detach() + step_size * torch.sign(grad.detach())
            x_adv = torch.min(
                torch.max(x_adv, x_natural - epsilon), x_natural + epsilon
            )
            x_adv = torch.clamp(x_adv, 0.0, 1.0)
    elif distance == "l_2":
        delta = 0.001 * torch.randn(x_natural.shape).cuda().detach()
        delta = Variable(delta.data, requires_grad=True)

        # Setup optimizers
        optimizer_delta = optim.SGD([delta], lr=epsilon / perturb_steps * 2)

        for _ in range(perturb_steps):
            adv = x_natural + delta

            # optimize
            optimizer_delta.zero_grad()
            with torch.enable_grad():
                loss = (-1) * criterion_kl(
                    F.log_softmax(model(adv), dim=1),
                    F.softmax(model(x_natural), dim=1),
                )
            loss.backward()
            # renorming gradient
            grad_norms = delta.grad.view(batch_size, -1).norm(p=2, dim=1)
            delta.grad.div_(grad_norms.view(-1, 1, 1, 1))
            # avoid nan or inf if gradient is 0
            if (grad_norms == 0).any():
                delta.grad[grad_norms == 0] = torch.randn_like(
                    delta.grad[grad_norms == 0]
                )
            optimizer_delta.step()

            # projection
            delta.data.add_(x_natural)
            delta.data.clamp_(0, 1).sub_(x_natural)
            delta.data.renorm_(p=2, dim=0, maxnorm=epsilon)
        x_adv = Variable(x_natural + delta, requires_grad=False)
    else:
        x_adv = torch.clamp(x_adv, 0.0, 1.0)
    model.train()

    x_adv = Variable(torch.clamp(x_adv, 0.0, 1.0), requires_grad=False)
    # zero gradient
    optimizer.zero_grad()
    # calculate robust loss
    logits = model(x_natural)
    loss_natural = F.cross_entropy(logits, y)
    """loss_robust = (1.0 / batch_size) * criterion_kl(F.log_softmax(model(x_adv), dim=1),
                                                    F.softmax(model(x_natural), dim=1))"""
    loss_robust = (1.0 / batch_size) * criterion_kl(
        F.log_softmax(model(x_adv), dim=1),
        torch.clamp(F.softmax(model(x_natural), dim=1), min=1e-8),
    )

    loss = loss_natural + beta * loss_robust
    return loss



from collections import OrderedDict
from datetime import datetime
import itertools

import numpy as np

import torch

import matplotlib.pyplot as plt
from ._vis import init_plot, make_twin, plot_line, _to_array


class HistoryMeter(object):
    def __init__(self, name):
        self.name = name
        self.history = {}
        self.val = None

    def __getitem__(self, idx):
        return self.history.get(idx)

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, self.name)

    def update(self, val, idx):
        self.history[idx] = val
        self.val = val

    def get_str(self):
        return None
    
    def state_dict(self):
        state_dict = OrderedDict(class_name=self.__class__.__name__,
                                 name=self.name,
                                 history=self.history,
                                 val=self.val,
                                )
        return state_dict
    
    def load_state_dict(self, state_dict):
        assert state_dict['class_name'] == self.__class__.__name__
        self.name = state_dict['name']
        self.history = state_dict['history']
        self.val = state_dict['val']
        return True


class IntMeter(HistoryMeter):
    def get_str(self):
        return "%d" % self.val

    def __getitem__(self, idx):
        val = self.history.get(idx)
        return val if val is not None else np.nan


class FloatMeter(HistoryMeter):
    def get_str(self):
        return "%.4f" % self.val

    def __getitem__(self, idx):
        val = self.history.get(idx)
        return val if val is not None else np.nan


class RecordManager(object):
    def __init__(self):
        self.records = OrderedDict()
        self.record_type = None
        self.save_path = None
        self.log = []  # Add logs since TextIOWrapper cannot saved.
        self.count = 0

        # For best records
        self.best_records = None
        self.best_option = None
        self._LBS = []
        self._HBS = []
        self._LBOS = []
        self._HBOS = []
        
        self._init_head = False
        self._init_time = datetime.now()
        self._progress_start_time = None
        self._progress_end_time = None
        self._progress_times = []

        self._spinner = itertools.cycle(["-", "/", "|", "\\"])
        
        self._print_form = None
        self._print_len = None

    def __getitem__(self, idx):
        return {key: meter.get(idx) for key, meter in self.records.items()}

    def __repr__(self):
        return "%s(keys=[%s])" % (
            self.__class__.__name__,
            ", ".join(self.records.keys()),
        )

    def initialize(self, record_type=None, save_path=None, best_option=None):
        self.record_type = record_type
        self.best_option = best_option
        self.save_path = save_path
        self._init_head = True
        if self.best_option is not None:
            self.define_best(self.best_option)

    def update(self, record_type=None, save_path=None, best_option=None):
        self.record_type = record_type
        self.best_option = best_option
        self.save_path = save_path
        self._init_head = True
        if self.best_option is not None:
            self.define_best(self.best_option)

    def progress_start(self):
        self._progress_start_time = datetime.now()

    def progress_end(self):
        self._progress_end_time = datetime.now()
        t = self._progress_end_time - self._progress_start_time
        self.print_only(self.record_type,
            "Progress: " + next(self._spinner) + " [" + str(t) + "/it]" + " " * 20,
            end="\r",
        )
        self._progress_times.append(t.total_seconds())

    def print(self, record_type, str, *args, **kargs):
        if record_type is not None:
            print(str, *args, **kargs)
            self.log.append(str)
            self.save_log()

    def print_only(self, record_type, str, *args, **kargs):
        if record_type is not None:
            print(str, *args, **kargs)

    def _add_progress_time(self, dict_record):
        dict_record["s/it"] = np.array(self._progress_times).mean()
        self._progress_times = []

    def add(self, dict_record):
        self.count += 1
        flag_new_key_addded = False
        dict_print = OrderedDict()
        self._add_progress_time(dict_record)

        for key, val in dict_record.items():
            # Check whether there exist meter for key.
            meter = self.records.get(key)
            if meter is None:
                val_type, val = self.check_type_and_transform(val)
                meter = self.assign_meter(val_type, key)
                self.records[key] = meter
                flag_new_key_addded = True
            meter.update(val, self.count)

            # If string, print.
            string = meter.get_str()
            if string:
                dict_print[key] = string

        if self.record_type == "Epoch":
            del dict_print["Iter"]

        # Print head
        if flag_new_key_addded or self._init_head:
            self._print_head(dict_print)

        # Print row
        self.print(self.record_type, self._print_form.format(*dict_print.values()))
        self.print(self.record_type, "-" * self._print_len)

    def _print_head(self, dict_print, slack=3):
        lengths = []
        for i, (key, value) in enumerate(dict_print.items()):
            length = max(len(value), len(key)) + slack
            lengths.append(length)

        self._print_form = "".join(
            ["{:<" + str(length) + "." + str(length) + "}" for length in lengths]
        )

        text = self._print_form.format(*dict_print.keys())
        self._print_len = len(text)
        self.print(self.record_type, "-" * self._print_len)
        self.print(self.record_type, text)
        self.print(self.record_type, "=" * self._print_len)
        self._init_head = False

    @staticmethod
    def assign_meter(val_type, key):
        if val_type == int:
            return IntMeter(key)
        elif val_type == float:
            return FloatMeter(key)
        return HistoryMeter(key)

    def define_best(self, best_option):
        for key, judge in best_option.items():
            if judge == "LB":
                self._LBS.append(key)
            elif judge == "HB":
                self._HBS.append(key)
            elif judge == "LBO":
                self._LBOS.append(key)
            elif judge == "HBO":
                self._HBOS.append(key)
            else:
                raise ValueError(
                    "Values of save_best should be in ['LB', 'HB' ,'LBO', 'HBO']."
                )

    def check_best(self, dict_record):
        if self.best_option is None:
            return False

        if dict_record is None:
            return False
        else:
            for key in self._LBS + self._HBS + self._LBOS + self._HBOS:
                if dict_record.get(key) is None:
                    return False

        if self.best_records is not None:
            flag_tie = True
            # Lower is better
            for key in self._LBS:
                best_value = self.best_records[key]
                curr_value = dict_record[key]
                if best_value < curr_value:
                    return False
                elif best_value > curr_value:
                    flag_tie = False
                else:
                    pass

            # Higher is better
            for key in self._HBS:
                best_value = self.best_records[key]
                curr_value = dict_record[key]
                if best_value > curr_value:
                    return False
                elif best_value < curr_value:
                    flag_tie = False
                else:
                    pass

            # If tie, go for options
            if flag_tie is True:
                # Lower is better
                for key in self._LBOS:
                    best_value = self.best_records[key]
                    curr_value = dict_record[key]
                    if best_value < curr_value:
                        return False

                # Higher is better
                for key in self._HBOS:
                    best_value = self.best_records[key]
                    curr_value = dict_record[key]
                    if best_value > curr_value:
                        return False
        self.best_records = dict_record
        return True

    def generate_summary(self):
        self.print(self.record_type, "=" * self._print_len)
        self.print(self.record_type, "Total Epoch: %d" % self.records["Epoch"].val)
        self.print(self.record_type, "Start Time: " + str(self._init_time))
        self.print(self.record_type, "Time Elapsed: " + str(datetime.now() - self._init_time))

        if self.best_option is not None:
            self.print(self.record_type, "Best Records: ")
            for key, val in self.best_records.items():
                if key in ["Epoch", "Iter"]:
                    self.print(self.record_type, "- %s: %d" % (key, val))
                if key in self.best_option.keys():
                    self.print(self.record_type, "- %s: %.4f" % (key, val))

        self.print(self.record_type, "-" * self._print_len)

    def save_log(self):
        if self.save_path is not None:
            with open(self.save_path + "/log.txt", "w") as f:
                f.write("\n".join(self.log) + "\n")

    def plot(
        self,
        inputs,
        x=None,
        figsize=(6, 6),
        title="",
        xlabel="",
        ylabel="",
        xlim=None,
        ylim=None,
        pad_ratio=0,
        tight=True,
        linestyles=None,
        linewidths=None,
        colors=None,
        labels=None,
        alphas=None,
        ylabel_second="",
        ylim_second=None,
        legend=True,
        loc="best",
    ):

        # Check version and number of elements
        ver = 1  # e.g.["a"] or ["a", "b", "c"]
        length = 0
        y_keys_flat = []
        js = []
        lines2 = []
        labels2 = []

        for j, y_key in enumerate(inputs):
            if isinstance(y_key, list):

                if len(inputs) > 2:
                    raise RuntimeError("Axes can have the maximum value as 2.")

                for y in y_key:
                    y_keys_flat.append(y)
                    length += 1
                    js.append(j)

                ver = 2  # e.g. [["a", "b"], "c"] or [["a", "b"], ["c", "d"]]

            else:
                y_keys_flat.append(y_key)
                length += 1
                js.append(j)

        idxs = list(range(self.count + 1))[1:]
        if x == "Epoch":
            v = self.records["Epoch"].history
            # Extract the last idx of each epoch.
            idxs = [idx for idx in idxs[:-1] if v[idx + 1] != v[idx]] + [idxs[-1]]
            x = [v[idx] for idx in idxs]
        else:
            x = idxs

        data = {}
        for key, meter in self.records.items():
            if isinstance(meter, IntMeter) or isinstance(meter, FloatMeter):
                data[key] = [meter.history[idx] for idx in idxs]

        inputs = [data[y_key] for y_key in y_keys_flat]

        # Draw plots
        ax = init_plot(
            ax=None,
            figsize=figsize,
            title=title,
            xlabel=xlabel,
            ylabel=ylabel,
            xlim=xlim,
            ylim=ylim,
            pad_ratio=pad_ratio,
            tight=tight,
        )

        linestyles, linewidths, colors, labels, alphas = _to_array(
            [linestyles, linewidths, colors, labels, alphas], length
        )

        i = 0
        if ver == 1:
            for input in inputs:
                plot_line(
                    ax,
                    input,
                    x,
                    linestyle=linestyles[i],
                    linewidth=linewidths[i],
                    color=colors[i],
                    label=labels[i],
                    alpha=alphas[i],
                )
                i += 1

        elif ver == 2:
            ax2 = make_twin(ax=ax, ylabel=ylabel_second, ylim=ylim_second)
            axes = [ax, ax2]
            for j, input in enumerate(inputs):
                plot_line(
                    axes[js[j]],
                    input,
                    x,
                    linestyle=linestyles[i],
                    linewidth=linewidths[i],
                    color=colors[i],
                    label=labels[i],
                    alpha=alphas[i],
                )
                i += 1

            lines2, labels2 = ax2.get_legend_handles_labels()

        else:
            raise RuntimeError("Unreadable inputs")

        if legend:
            lines, labels = ax.get_legend_handles_labels()
            ax.legend(lines + lines2, labels + labels2, loc=loc)

        ax.set_xticks(x)
        plt.show()

    @staticmethod
    def check_type_and_transform(val):
        # if isinstance(val, torch.Tensor):
        #     if val.numel() == 1:
        #         val = val.item()
        # if isinstance(val, np.ndarray):
        #     if val.size == 1:
        # Instead above,
        try:
            val = val.item()
        except:
            pass
        return type(val), val

    def state_dict(self):
        records = OrderedDict()
        for key in self.records.keys():
            records[key] = self.records[key].state_dict()
        
        state_dict = OrderedDict(class_name=self.__class__.__name__,
                                 records=records,
                                 record_type=self.record_type,
                                 best_records=self.best_records,
                                 best_option=self.best_option,
                                 log=self.log,
                                 count=self.count,
                                 _init_head=self._init_head,
                                 _init_time=self._init_time,
                                 _progress_start_time=self._progress_start_time,
                                 _progress_end_time=self._progress_end_time,
                                 _progress_times=self._progress_times,
                                 _spinner=self._spinner,
                                 save_path=self.save_path,
                                 _LBS=self._LBS,
                                 _HBS=self._HBS,
                                 _LBOS=self._LBOS,
                                 _HBOS=self._HBOS,
                                )
        return state_dict
        
    def load_state_dict(self, state_dict):
        assert state_dict['class_name'] == self.__class__.__name__
        for key in state_dict.keys():
            if key in ['class_name']:
                continue
            elif key in ['records']:
                records = OrderedDict()
                records_dict = state_dict['records']
                for record_key in records_dict.keys():
                    records[record_key] = globals()[records_dict[record_key]['class_name']](None)
                    records[record_key].load_state_dict(records_dict[record_key])
                self.records = records
            else:
                setattr(self, key, state_dict[key])
        return True


import torch
import torch.nn as nn

from ..attack import Attack


class PGDL2(Attack):
    r"""
    PGD in the paper 'Towards Deep Learning Models Resistant to Adversarial Attacks'
    [https://arxiv.org/abs/1706.06083]

    Distance Measure : L2

    Arguments:
        model (nn.Module): model to attack.
        eps (float): maximum perturbation. (Default: 1.0)
        alpha (float): step size. (Default: 0.2)
        steps (int): number of steps. (Default: 10)
        random_start (bool): using random initialization of delta. (Default: True)

    Shape:
        - images: :math:`(N, C, H, W)` where `N = number of batches`, `C = number of channels`,        `H = height` and `W = width`. It must have a range [0, 1].
        - labels: :math:`(N)` where each value :math:`y_i` is :math:`0 \leq y_i \leq` `number of labels`.
        - output: :math:`(N, C, H, W)`.

    Examples::
        >>> attack = torchattacks.PGDL2(model, eps=1.0, alpha=0.2, steps=10, random_start=True)
        >>> adv_images = attack(images, labels)

    """

    def __init__(
        self,
        model,
        eps=1.0,
        alpha=0.2,
        steps=10,
        random_start=True,
        eps_for_division=1e-10,
    ):
        super().__init__("PGDL2", model)
        self.eps = eps
        self.alpha = alpha
        self.steps = steps
        self.random_start = random_start
        self.eps_for_division = eps_for_division
        self.supported_mode = ["default", "targeted"]

    def forward(self, images, labels):
        r"""
        Overridden.
        """

        images = images.clone().detach().to(self.device)
        labels = labels.clone().detach().to(self.device)

        if self.targeted:
            target_labels = self.get_target_label(images, labels)

        loss = nn.CrossEntropyLoss()

        adv_images = images.clone().detach()
        batch_size = len(images)

        if self.random_start:
            # Starting at a uniformly random point
            delta = torch.empty_like(adv_images).normal_()
            d_flat = delta.view(adv_images.size(0), -1)
            n = d_flat.norm(p=2, dim=1).view(adv_images.size(0), 1, 1, 1)
            r = torch.zeros_like(n).uniform_(0, 1)
            delta *= r / n * self.eps
            adv_images = torch.clamp(adv_images + delta, min=0, max=1).detach()

        for _ in range(self.steps):
            adv_images.requires_grad = True
            outputs = self.get_logits(adv_images)

            # Calculate loss
            if self.targeted:
                cost = -loss(outputs, target_labels)
            else:
                cost = loss(outputs, labels)

            # Update adversarial images
            grad = torch.autograd.grad(
                cost, adv_images, retain_graph=False, create_graph=False
            )[0]
            grad_norms = (
                torch.norm(grad.view(batch_size, -1), p=2, dim=1)
                + self.eps_for_division
            )  # nopep8
            grad = grad / grad_norms.view(batch_size, 1, 1, 1)
            adv_images = adv_images.detach() + self.alpha * grad

            delta = adv_images - images
            delta_norms = torch.norm(delta.view(batch_size, -1), p=2, dim=1)
            factor = self.eps / delta_norms
            factor = torch.min(factor, torch.ones_like(delta_norms))
            delta = delta * factor.view(-1, 1, 1, 1)

            adv_images = torch.clamp(images + delta, min=0, max=1).detach()

        return adv_images
